import logging
import os
import re
from pathlib import Path
from urllib.request import urlopen

import pytest

from rdflib import Graph, Literal, URIRef
from rdflib.plugins.parsers import ntriples
from test.data import TEST_DATA_DIR

log = logging.getLogger(__name__)

NT_PATH = os.path.relpath(os.path.join(TEST_DATA_DIR, "suites", "nt_misc"), os.curdir)


def nt_file(fn):
    return os.path.join(NT_PATH, fn)


# Test NT


def test_issue859():
    graph_a = Graph()
    graph_b = Graph()
    graph_a.parse(nt_file("quote-01.nt"), format="ntriples")
    graph_b.parse(nt_file("quote-02.nt"), format="ntriples")
    for subject_a, predicate_a, obj_a in graph_a:
        for subject_b, predicate_b, obj_b in graph_b:
            assert subject_a == subject_b
            assert predicate_a == predicate_b
            assert obj_a == obj_b


def test_issue78():
    g = Graph()
    g.add((URIRef("foo"), URIRef("foo"), Literal("R\u00E4ksm\u00F6rg\u00E5s")))
    s = g.serialize(format="nt")
    assert type(s) == str  # noqa: E721
    assert "R\u00E4ksm\u00F6rg\u00E5s" in s


def test_issue146():
    g = Graph()
    g.add((URIRef("foo"), URIRef("foo"), Literal("test\n", lang="en")))
    s = g.serialize(format="nt").strip()
    assert s == '<foo> <foo> "test\\n"@en .'


def test_issue1144_rdflib():
    fname = nt_file("lists-02.nt")
    g1 = Graph()
    with open(fname, "r") as f:
        g1.parse(f, format="nt")
    assert 14 == len(g1)
    g2 = Graph()
    with open(fname, "rb") as fb:
        g2.parse(fb, format="nt")
    assert 14 == len(g2)


def test_issue1144_w3c():
    fname = nt_file("lists-02.nt")
    sink1 = ntriples.NTGraphSink(Graph())
    p1 = ntriples.W3CNTriplesParser(sink1)
    with open(fname, "r") as f:
        p1.parse(f)
    assert 14 == len(sink1.g)
    sink2 = ntriples.NTGraphSink(Graph())
    p2 = ntriples.W3CNTriplesParser(sink2)
    with open(fname, "rb") as f:
        p2.parse(f)
    assert 14 == len(sink2.g)


def test_sink():
    s = ntriples.DummySink()
    assert s.length == 0
    s.triple(None, None, None)
    assert s.length == 1


def test_nonvalidating_unquote():
    safe = """<http://example.org/alice/foaf.rdf#me> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Person> <http://example.org/alice/foaf1.rdf> ."""
    ntriples.validate = False
    res = ntriples.unquote(safe)
    assert isinstance(res, str)


def test_validating_unquote():
    quot = """<http://example.org/alice/foaf.rdf#me> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Person> <http://example.org/alice/foaf1.rdf> ."""
    ntriples.validate = True
    res = ntriples.unquote(quot)
    # revert to default
    ntriples.validate = False
    log.debug("restype %s" % type(res))


def test_validating_unquote_raises():
    ntriples.validate = True
    uniquot = """<http://www.w3.org/People/Berners-Lee/card#cm> <http://xmlns.com/foaf/0.1/name> "R\\u00E4ksm\\u00F6rg\\u00E5s" <http://www.w3.org/People/Berners-Lee/card> ."""
    with pytest.raises(ntriples.ParseError):
        ntriples.unquote(uniquot)
    uniquot = """<http://www.w3.org/People/Berners-Lee/card#cm> <http://xmlns.com/foaf/0.1/name> "R\\\\u00E4ksm\\u00F6rg\\u00E5s" <http://www.w3.org/People/Berners-Lee/card> ."""
    with pytest.raises(ntriples.ParseError):
        ntriples.unquote(uniquot)
    # revert to default
    ntriples.validate = False


def test_nonvalidating_uriquote():
    ntriples.validate = False
    safe = """<http://example.org/alice/foaf.rdf#me> <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://xmlns.com/foaf/0.1/Person> <http://example.org/alice/foaf1.rdf> ."""
    res = ntriples.uriquote(safe)
    assert res == safe


def test_validating_uriquote():
    ntriples.validate = True
    uniquot = """<http://www.w3.org/People/Berners-Lee/card#cm> <http://xmlns.com/foaf/0.1/name> "R\\u00E4ksm\\u00F6rg\\u00E5s" <http://www.w3.org/People/Berners-Lee/card> ."""
    res = ntriples.uriquote(uniquot)
    # revert to default
    ntriples.validate = False
    assert res == uniquot


def test_w3d_ntriples_parser_fpath():
    fpath = os.path.join(nt_file(os.listdir(NT_PATH)[0]))
    p = ntriples.W3CNTriplesParser()
    with pytest.raises(ntriples.ParseError):
        p.parse(fpath)


def test_w3c_ntriples_parser_parsestring():
    p = ntriples.W3CNTriplesParser()
    data = 3
    with pytest.raises(ntriples.ParseError):
        p.parsestring(data)
    with open(nt_file("lists-02.nt"), "r") as f:
        data = f.read()
    p = ntriples.W3CNTriplesParser()
    res = p.parsestring(data)
    assert res is None


def test_w3_ntriple_variants():
    uri = Path(nt_file("test.nt")).absolute().as_uri()

    parser = ntriples.W3CNTriplesParser()
    u = urlopen(uri)
    sink = parser.parse(u)
    u.close()
    # ATM we are only really interested in any exceptions thrown
    assert sink is not None


def test_bad_line():
    data = """<http://example.org/resource32> 3 <http://example.org/datatype1> .\n"""
    p = ntriples.W3CNTriplesParser()
    with pytest.raises(ntriples.ParseError):
        p.parsestring(data)


def test_cover_eat():
    data = """<http://example.org/resource32> 3 <http://example.org/datatype1> .\n"""
    p = ntriples.W3CNTriplesParser()
    p.line = data
    with pytest.raises(ntriples.ParseError):
        p.eat(re.compile("<http://example.org/datatype1>"))


def test_cover_subjectobjectliteral():
    # data = '''<http://example.org/resource32> 3 <http://example.org/datatype1> .\n'''
    p = ntriples.W3CNTriplesParser()
    p.line = "baz"
    with pytest.raises(ntriples.ParseError):
        p.subject()
    with pytest.raises(ntriples.ParseError):
        p.object()
    # p.line = '"baz"@fr^^<http://example.org/datatype1>'
    # self.assertRaises(ntriples.ParseError, p.literal)


# Test BNode context


def test_bnode_shared_across_instances():
    my_sink = FakeSink()
    bnode_context = dict()
    p = ntriples.W3CNTriplesParser(my_sink, bnode_context=bnode_context)
    p.parsestring(
        """
    _:0 <http://purl.obolibrary.org/obo/RO_0002350> <http://www.gbif.org/species/0000001> .
    """
    )

    q = ntriples.W3CNTriplesParser(my_sink, bnode_context=bnode_context)
    q.parsestring(
        """
    _:0 <http://purl.obolibrary.org/obo/RO_0002350> <http://www.gbif.org/species/0000002> .
    """
    )

    assert len(my_sink.subs) == 1


def test_bnode_distinct_across_instances():
    my_sink = FakeSink()
    p = ntriples.W3CNTriplesParser(my_sink)
    p.parsestring(
        """
    _:0 <http://purl.obolibrary.org/obo/RO_0002350> <http://www.gbif.org/species/0000001> .
    """
    )

    q = ntriples.W3CNTriplesParser(my_sink)
    q.parsestring(
        """
    _:0 <http://purl.obolibrary.org/obo/RO_0002350> <http://www.gbif.org/species/0000002> .
    """
    )

    assert len(my_sink.subs) == 2


def test_bnode_distinct_across_parse():
    my_sink = FakeSink()
    p = ntriples.W3CNTriplesParser(my_sink)

    p.parsestring(
        """
    _:0 <http://purl.obolibrary.org/obo/RO_0002350> <http://www.gbif.org/species/0000001> .
    """,
        bnode_context=dict(),
    )

    p.parsestring(
        """
    _:0 <http://purl.obolibrary.org/obo/RO_0002350> <http://www.gbif.org/species/0000002> .
    """,
        bnode_context=dict(),
    )

    assert len(my_sink.subs) == 2


def test_bnode_shared_across_parse():
    my_sink = FakeSink()
    p = ntriples.W3CNTriplesParser(my_sink)

    p.parsestring(
        """
    _:0 <http://purl.obolibrary.org/obo/RO_0002350> <http://www.gbif.org/species/0000001> .
    """
    )

    p.parsestring(
        """
    _:0 <http://purl.obolibrary.org/obo/RO_0002350> <http://www.gbif.org/species/0000002> .
    """
    )

    assert len(my_sink.subs) == 1


def test_bnode_shared_across_instances_with_parse_option():
    my_sink = FakeSink()
    bnode_ctx = dict()

    p = ntriples.W3CNTriplesParser(my_sink)
    p.parsestring(
        """
    _:0 <http://purl.obolibrary.org/obo/RO_0002350> <http://www.gbif.org/species/0000001> .
    """,
        bnode_context=bnode_ctx,
    )

    q = ntriples.W3CNTriplesParser(my_sink)
    q.parsestring(
        """
    _:0 <http://purl.obolibrary.org/obo/RO_0002350> <http://www.gbif.org/species/0000002> .
    """,
        bnode_context=bnode_ctx,
    )

    assert len(my_sink.subs) == 1


class FakeSink:
    def __init__(self):
        self.subs = set()

    def triple(self, s, p, o):
        self.subs.add(s)



import pytest

from rdflib import RDFS, Graph, Literal, URIRef


@pytest.fixture(scope="function")
def graph() -> Graph:
    graph = Graph()
    # Bind prefixes.
    graph.bind("isbn", "urn:isbn:")
    graph.bind("webn", "http://w3c.org/example/isbn/")
    # Populate graph.
    graph.add((URIRef("urn:isbn:1503280780"), RDFS.label, Literal("Moby Dick")))
    graph.add(
        (
            URIRef("http://w3c.org/example/isbn/1503280780"),
            RDFS.label,
            Literal("Moby Dick"),
        )
    )
    return graph


def test_namespace_manager(graph: Graph):
    assert "isbn", "urn:isbn:" in tuple(graph.namespaces())
    assert "webn", "http://w3c.org/example/isbn/" in tuple(graph.namespaces())


def test_turtle_serialization(graph: Graph):
    serialization = graph.serialize(None, format="turtle")
    print(f"Test Issue 977, serialization output:\n---\n{serialization}---")
    # Test serialization.
    assert "@prefix webn:" in serialization, "Prefix webn not found in serialization!"
    assert "@prefix isbn:" in serialization, "Prefix isbn not found in serialization!"



from rdflib import Dataset, Graph, Literal, URIRef


def main():
    # example for adding a quad
    ds = Dataset()
    g = Graph(identifier=URIRef("http://graph-a"))
    ds.add_graph(g)
    triple = (URIRef("http://subj-a"), URIRef("http://pred-a"), Literal("obj-a"))
    ds.get_context(g.identifier).add(triple)
    result = ds.serialize(format="patch", operation="add")
    print("Add Quad Patch:")
    print(result)

    # alternate example for adding a quad
    ds = Dataset()
    quad = (
        URIRef("http://subj-a"),
        URIRef("http://pred-a"),
        Literal("obj-a"),
        Graph(identifier=URIRef("http://graph-a")),
    )
    ds.add(quad)
    result = ds.serialize(format="patch", operation="add")
    print("Add Quad Patch:")
    print(result)

    # example for adding a triple
    ds = Dataset()
    ds.add(triple)
    result = ds.serialize(format="patch", operation="add")
    print("\nAdd Triple Patch:")
    print(result)

    # Example for diff quads
    quad_1 = (
        URIRef("http://subj-a"),
        URIRef("http://pred-a"),
        Literal("obj-a"),
        Graph(identifier=URIRef("http://graph-a")),
    )
    quad_2 = (
        URIRef("http://subj-b"),
        URIRef("http://pred-b"),
        Literal("obj-b"),
        Graph(identifier=URIRef("http://graph-b")),
    )
    quad_3 = (
        URIRef("http://subj-c"),
        URIRef("http://pred-c"),
        Literal("obj-c"),
        Graph(identifier=URIRef("http://graph-c")),
    )
    ds1 = Dataset()
    ds2 = Dataset()
    ds1.addN([quad_1, quad_2])
    ds2.addN([quad_2, quad_3])
    result = ds1.serialize(format="patch", target=ds2)
    print("Diff Quad Patch:")
    print(result)


if __name__ == "__main__":
    main()



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class DCTERMS(DefinedNamespace):
    """
    DCMI Metadata Terms - other

    Generated from: https://www.dublincore.org/specifications/dublin-core/dcmi-terms/dublin_core_terms.ttl
    Date: 2020-05-26 14:20:00.590514

    """

    _fail = True

    # http://purl.org/dc/dcam/VocabularyEncodingScheme
    DCMIType: URIRef  # The set of classes specified by the DCMI Type Vocabulary, used to categorize the nature or genre of the resource.
    DDC: URIRef  # The set of conceptual resources specified by the Dewey Decimal Classification.
    IMT: URIRef  # The set of media types specified by the Internet Assigned Numbers Authority.
    LCC: URIRef  # The set of conceptual resources specified by the Library of Congress Classification.
    LCSH: URIRef  # The set of labeled concepts specified by the Library of Congress Subject Headings.
    MESH: (
        URIRef  # The set of labeled concepts specified by the Medical Subject Headings.
    )
    NLM: URIRef  # The set of conceptual resources specified by the National Library of Medicine Classification.
    TGN: URIRef  # The set of places specified by the Getty Thesaurus of Geographic Names.
    UDC: URIRef  # The set of conceptual resources specified by the Universal Decimal Classification.

    # http://www.w3.org/1999/02/22-rdf-syntax-ns#Property
    abstract: URIRef  # A summary of the resource.
    accessRights: URIRef  # Information about who access the resource or an indication of its security status.
    accrualMethod: URIRef  # The method by which items are added to a collection.
    accrualPeriodicity: (
        URIRef  # The frequency with which items are added to a collection.
    )
    accrualPolicy: URIRef  # The policy governing the addition of items to a collection.
    alternative: URIRef  # An alternative name for the resource.
    audience: URIRef  # A class of agents for whom the resource is intended or useful.
    available: URIRef  # Date that the resource became or will become available.
    bibliographicCitation: URIRef  # A bibliographic reference for the resource.
    conformsTo: (
        URIRef  # An established standard to which the described resource conforms.
    )
    contributor: (
        URIRef  # An entity responsible for making contributions to the resource.
    )
    coverage: URIRef  # The spatial or temporal topic of the resource, spatial applicability of the resource, or jurisdiction under which the resource is relevant.
    created: URIRef  # Date of creation of the resource.
    creator: URIRef  # An entity responsible for making the resource.
    date: URIRef  # A point or period of time associated with an event in the lifecycle of the resource.
    dateAccepted: URIRef  # Date of acceptance of the resource.
    dateCopyrighted: URIRef  # Date of copyright of the resource.
    dateSubmitted: URIRef  # Date of submission of the resource.
    description: URIRef  # An account of the resource.
    educationLevel: URIRef  # A class of agents, defined in terms of progression through an educational or training context, for which the described resource is intended.
    extent: URIRef  # The size or duration of the resource.
    format: URIRef  # The file format, physical medium, or dimensions of the resource.
    hasFormat: URIRef  # A related resource that is substantially the same as the pre-existing described resource, but in another format.
    hasPart: URIRef  # A related resource that is included either physically or logically in the described resource.
    hasVersion: URIRef  # A related resource that is a version, edition, or adaptation of the described resource.
    identifier: (
        URIRef  # An unambiguous reference to the resource within a given context.
    )
    instructionalMethod: URIRef  # A process, used to engender knowledge, attitudes and skills, that the described resource is designed to support.
    isFormatOf: URIRef  # A pre-existing related resource that is substantially the same as the described resource, but in another format.
    isPartOf: URIRef  # A related resource in which the described resource is physically or logically included.
    isReferencedBy: URIRef  # A related resource that references, cites, or otherwise points to the described resource.
    isReplacedBy: URIRef  # A related resource that supplants, displaces, or supersedes the described resource.
    isRequiredBy: URIRef  # A related resource that requires the described resource to support its function, delivery, or coherence.
    isVersionOf: URIRef  # A related resource of which the described resource is a version, edition, or adaptation.
    issued: URIRef  # Date of formal issuance of the resource.
    language: URIRef  # A language of the resource.
    license: URIRef  # A legal document giving official permission to do something with the resource.
    mediator: URIRef  # An entity that mediates access to the resource.
    medium: URIRef  # The material or physical carrier of the resource.
    modified: URIRef  # Date on which the resource was changed.
    provenance: URIRef  # A statement of any changes in ownership and custody of the resource since its creation that are significant for its authenticity, integrity, and interpretation.
    publisher: URIRef  # An entity responsible for making the resource available.
    references: URIRef  # A related resource that is referenced, cited, or otherwise pointed to by the described resource.
    relation: URIRef  # A related resource.
    replaces: URIRef  # A related resource that is supplanted, displaced, or superseded by the described resource.
    requires: URIRef  # A related resource that is required by the described resource to support its function, delivery, or coherence.
    rights: URIRef  # Information about rights held in and over the resource.
    rightsHolder: (
        URIRef  # A person or organization owning or managing rights over the resource.
    )
    source: URIRef  # A related resource from which the described resource is derived.
    spatial: URIRef  # Spatial characteristics of the resource.
    subject: URIRef  # A topic of the resource.
    tableOfContents: URIRef  # A list of subunits of the resource.
    temporal: URIRef  # Temporal characteristics of the resource.
    title: URIRef  # A name given to the resource.
    type: URIRef  # The nature or genre of the resource.
    valid: URIRef  # Date (often a range) of validity of a resource.

    # http://www.w3.org/2000/01/rdf-schema#Class
    Agent: URIRef  # A resource that acts or has the power to act.
    AgentClass: URIRef  # A group of agents.
    BibliographicResource: URIRef  # A book, article, or other documentary resource.
    FileFormat: URIRef  # A digital resource format.
    Frequency: URIRef  # A rate at which something recurs.
    Jurisdiction: (
        URIRef  # The extent or range of judicial, law enforcement, or other authority.
    )
    LicenseDocument: URIRef  # A legal document giving official permission to do something with a resource.
    LinguisticSystem: URIRef  # A system of signs, symbols, sounds, gestures, or rules used in communication.
    Location: URIRef  # A spatial region or named place.
    LocationPeriodOrJurisdiction: URIRef  # A location, period of time, or jurisdiction.
    MediaType: URIRef  # A file format or physical medium.
    MediaTypeOrExtent: URIRef  # A media type or extent.
    MethodOfAccrual: URIRef  # A method by which resources are added to a collection.
    MethodOfInstruction: (
        URIRef  # A process that is used to engender knowledge, attitudes, and skills.
    )
    PeriodOfTime: URIRef  # An interval of time that is named or defined by its start and end dates.
    PhysicalMedium: URIRef  # A physical material or carrier.
    PhysicalResource: URIRef  # A material thing.
    Policy: URIRef  # A plan or course of action by an authority, intended to influence and determine decisions, actions, and other matters.
    ProvenanceStatement: URIRef  # Any changes in ownership and custody of a resource since its creation that are significant for its authenticity, integrity, and interpretation.
    RightsStatement: URIRef  # A statement about the intellectual property rights (IPR) held in or over a resource, a legal document giving official permission to do something with a resource, or a statement about access rights.
    SizeOrDuration: URIRef  # A dimension or extent, or a time taken to play or execute.
    Standard: URIRef  # A reference point against which other things can be evaluated or compared.

    # http://www.w3.org/2000/01/rdf-schema#Datatype
    Box: URIRef  # The set of regions in space defined by their geographic coordinates according to the DCMI Box Encoding Scheme.
    ISO3166: URIRef  # The set of codes listed in ISO 3166-1 for the representation of names of countries.
    Period: URIRef  # The set of time intervals defined by their limits according to the DCMI Period Encoding Scheme.
    Point: URIRef  # The set of points in space defined by their geographic coordinates according to the DCMI Point Encoding Scheme.
    RFC1766: URIRef  # The set of tags, constructed according to RFC 1766, for the identification of languages.
    RFC3066: URIRef  # The set of tags constructed according to RFC 3066 for the identification of languages.
    RFC4646: URIRef  # The set of tags constructed according to RFC 4646 for the identification of languages.
    RFC5646: URIRef  # The set of tags constructed according to RFC 5646 for the identification of languages.
    URI: URIRef  # The set of identifiers constructed according to the generic syntax for Uniform Resource Identifiers as specified by the Internet Engineering Task Force.
    W3CDTF: URIRef  # The set of dates and times constructed according to the W3C Date and Time Formats Specification.

    # Valid non-python identifiers
    _extras = ["ISO639-2", "ISO639-3"]

    _NS = Namespace("http://purl.org/dc/terms/")



from __future__ import annotations

from typing import Tuple
from urllib.parse import urlparse

from rdflib.graph import Graph
from test.data import SIMPLE_TRIPLE_GRAPH, TEST_DATA_DIR
from test.utils import GraphHelper
from test.utils.http import MethodName, MockHTTPResponse
from test.utils.httpservermock import ServedBaseHTTPServerMock


def test_graph_redirect_new_host(
    function_httpmocks: Tuple[ServedBaseHTTPServerMock, ServedBaseHTTPServerMock]
) -> None:
    """
    Redirect to new host results in a request with the right Host header
    parameter.
    """

    mock_a, mock_b = function_httpmocks

    mock_a.responses[MethodName.GET].append(
        MockHTTPResponse(
            308,
            "Permanent Redirect",
            b"",
            {"Location": [f"{mock_b.url}/b/data.ttl"]},
        )
    )

    mock_b.responses[MethodName.GET].append(
        MockHTTPResponse(
            200,
            "OK",
            (TEST_DATA_DIR / "variants" / "simple_triple.ttl").read_bytes(),
            {"Content-Type": ["text/turtle"]},
        )
    )

    graph = Graph()
    graph.parse(location=f"{mock_a.url}/a/data.ttl")
    GraphHelper.assert_sets_equals(graph, SIMPLE_TRIPLE_GRAPH)
    for mock in function_httpmocks:
        assert 1 == len(mock.requests[MethodName.GET])
        for request in mock.requests[MethodName.GET]:
            assert request.headers["Host"] == urlparse(mock.url).netloc



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class PROV(DefinedNamespace):
    """
    W3C PROVenance Interchange Ontology (PROV-O)

    This document is published by the Provenance Working Group (http://www.w3.org/2011/prov/wiki/Main_Page).   If
    you wish to make comments regarding this document, please send them to public-prov-comments@w3.org (subscribe
    public-prov-comments-request@w3.org, archives http://lists.w3.org/Archives/Public/public-prov-comments/). All
    feedback is welcome.

    PROV Access and Query Ontology

    This document is published by the Provenance Working Group (http://www.w3.org/2011/prov/wiki/Main_Page).   If
    you wish to make comments regarding this document, please send them to public-prov-comments@w3.org (subscribe
    public-prov-comments-request@w3.org, archives http://lists.w3.org/Archives/Public/public-prov-comments/). All
    feedback is welcome.

    Dublin Core extensions of the W3C PROVenance Interchange Ontology (PROV-O)

    This document is published by the Provenance Working Group (http://www.w3.org/2011/prov/wiki/Main_Page).   If
    you wish to make comments regarding this document, please send them to public-prov-comments@w3.org (subscribe
    public-prov-comments-request@w3.org, archives http://lists.w3.org/Archives/Public/public-prov-comments/). All
    feedback is welcome.

    W3C PROV Linking Across Provenance Bundles Ontology (PROV-LINKS)

    This document is published by the Provenance Working Group (http://www.w3.org/2011/prov/wiki/Main_Page). If
    you wish to make comments regarding this document, please send them to public-prov-comments@w3.org (subscribe
    public-prov-comments-request@w3.org, archives http://lists.w3.org/Archives/Public/public-prov-comments/ ). All
    feedback is welcome.

    W3C PROVenance Interchange Ontology (PROV-O) Dictionary Extension

    This document is published by the Provenance Working Group (http://www.w3.org/2011/prov/wiki/Main_Page).
    If you wish to make comments regarding this document, please send them to public-prov-comments@w3.org
    (subscribe public-prov-comments-request@w3.org, archives http://lists.w3.org/Archives/Public/public-prov-
    comments/). All feedback is welcome.

    W3C PROVenance Interchange

    This document is published by the Provenance Working Group (http://www.w3.org/2011/prov/wiki/Main_Page).  If
    you wish to make comments regarding this document, please send them to public-prov-comments@w3.org (subscribe
    public-prov-comments-request@w3.org, archives http://lists.w3.org/ Archives/Public/public-prov-comments/). All
    feedback is welcome.

    Generated from: http://www.w3.org/ns/prov
    Date: 2020-05-26 14:20:04.650279

    """

    _fail = True

    # http://www.w3.org/2000/01/rdf-schema#Resource
    activityOfInfluence: URIRef  # activityOfInfluence
    agentOfInfluence: URIRef  # agentOfInfluence
    contributed: URIRef  # contributed
    ended: URIRef  # ended
    entityOfInfluence: URIRef  # entityOfInfluence
    generalizationOf: URIRef  # generalizationOf
    generatedAsDerivation: URIRef  # generatedAsDerivation
    hadDelegate: URIRef  # hadDelegate
    hadDerivation: URIRef  # hadDerivation
    hadInfluence: URIRef  # hadInfluence
    hadRevision: URIRef  # hadRevision
    informed: URIRef  # informed
    locationOf: URIRef  # locationOf
    qualifiedAssociationOf: URIRef  # qualifiedAssociationOf
    qualifiedAttributionOf: URIRef  # qualifiedAttributionOf
    qualifiedCommunicationOf: URIRef  # qualifiedCommunicationOf
    qualifiedDelegationOf: URIRef  # qualifiedDelegationOf
    qualifiedDerivationOf: URIRef  # qualifiedDerivationOf
    qualifiedEndOf: URIRef  # qualifiedEndOf
    qualifiedGenerationOf: URIRef  # qualifiedGenerationOf
    qualifiedInfluenceOf: URIRef  # qualifiedInfluenceOf
    qualifiedInvalidationOf: URIRef  # qualifiedInvalidationOf
    qualifiedQuotationOf: URIRef  # qualifiedQuotationOf
    qualifiedSourceOf: URIRef  # qualifiedSourceOf
    qualifiedStartOf: URIRef  # qualifiedStartOf
    qualifiedUsingActivity: URIRef  # qualifiedUsingActivity
    quotedAs: URIRef  # quotedAs
    revisedEntity: URIRef  # revisedEntity
    started: URIRef  # started
    wasActivityOfInfluence: URIRef  # wasActivityOfInfluence
    wasAssociateFor: URIRef  # wasAssociateFor
    wasMemberOf: URIRef  # wasMemberOf
    wasPlanOf: URIRef  # wasPlanOf
    wasPrimarySourceOf: URIRef  # wasPrimarySourceOf
    wasRoleIn: URIRef  # wasRoleIn
    wasUsedBy: URIRef  # wasUsedBy
    wasUsedInDerivation: URIRef  # wasUsedInDerivation

    # http://www.w3.org/2002/07/owl#AnnotationProperty
    aq: URIRef  #
    category: URIRef  # Classify prov-o terms into three categories, including 'starting-point', 'qualifed', and 'extended'. This classification is used by the prov-o html document to gently introduce prov-o terms to its users.
    component: URIRef  # Classify prov-o terms into six components according to prov-dm, including 'agents-responsibility', 'alternate', 'annotations', 'collections', 'derivations', and 'entities-activities'. This classification is used so that readers of prov-o specification can find its correspondence with the prov-dm specification.
    constraints: URIRef  # A reference to the principal section of the PROV-CONSTRAINTS document that describes this concept.
    definition: URIRef  # A definition quoted from PROV-DM or PROV-CONSTRAINTS that describes the concept expressed with this OWL term.
    dm: URIRef  # A reference to the principal section of the PROV-DM document that describes this concept.
    editorialNote: URIRef  # A note by the OWL development team about how this term expresses the PROV-DM concept, or how it should be used in context of semantic web or linked data.
    editorsDefinition: URIRef  # When the prov-o term does not have a definition drawn from prov-dm, and the prov-o editor provides one.
    inverse: URIRef  # PROV-O does not define all property inverses. The directionalities defined in PROV-O should be given preference over those not defined. However, if users wish to name the inverse of a PROV-O property, the local name given by prov:inverse should be used.
    n: URIRef  # A reference to the principal section of the PROV-M document that describes this concept.
    order: URIRef  # The position that this OWL term should be listed within documentation. The scope of the documentation (e.g., among all terms, among terms within a prov:category, among properties applying to a particular class, etc.) is unspecified.
    qualifiedForm: URIRef  # This annotation property links a subproperty of prov:wasInfluencedBy with the subclass of prov:Influence and the qualifying property that are used to qualify it.   Example annotation:      prov:wasGeneratedBy prov:qualifiedForm prov:qualifiedGeneration, prov:Generation .  Then this unqualified assertion:      :entity1 prov:wasGeneratedBy :activity1 .  can be qualified by adding:     :entity1 prov:qualifiedGeneration :entity1Gen .    :entity1Gen         a prov:Generation, prov:Influence;        prov:activity :activity1;        :customValue 1337 .  Note how the value of the unqualified influence (prov:wasGeneratedBy :activity1) is mirrored as the value of the prov:activity (or prov:entity, or prov:agent) property on the influence class.
    sharesDefinitionWith: URIRef  #
    specializationOf: URIRef  # specializationOf
    todo: URIRef  #
    unqualifiedForm: URIRef  # Classes and properties used to qualify relationships are annotated with prov:unqualifiedForm to indicate the property used to assert an unqualified provenance relation.
    wasRevisionOf: URIRef  # A revision is a derivation that revises an entity into a revised version.

    # http://www.w3.org/2002/07/owl#Class
    Accept: URIRef  # Accept
    Activity: URIRef  # Activity
    ActivityInfluence: URIRef  # ActivityInfluence provides additional descriptions of an Activity's binary influence upon any other kind of resource. Instances of ActivityInfluence use the prov:activity property to cite the influencing Activity.
    Agent: URIRef  # Agent
    AgentInfluence: URIRef  # AgentInfluence provides additional descriptions of an Agent's binary influence upon any other kind of resource. Instances of AgentInfluence use the prov:agent property to cite the influencing Agent.
    Association: URIRef  # An instance of prov:Association provides additional descriptions about the binary prov:wasAssociatedWith relation from an prov:Activity to some prov:Agent that had some responsibility for it. For example, :baking prov:wasAssociatedWith :baker; prov:qualifiedAssociation [ a prov:Association; prov:agent :baker; :foo :bar ].
    Attribution: URIRef  # An instance of prov:Attribution provides additional descriptions about the binary prov:wasAttributedTo relation from an prov:Entity to some prov:Agent that had some responsible for it. For example, :cake prov:wasAttributedTo :baker; prov:qualifiedAttribution [ a prov:Attribution; prov:entity :baker; :foo :bar ].
    Bundle: URIRef  # Note that there are kinds of bundles (e.g. handwritten letters, audio recordings, etc.) that are not expressed in PROV-O, but can be still be described by PROV-O.
    Collection: URIRef  # Collection
    Communication: URIRef  # An instance of prov:Communication provides additional descriptions about the binary prov:wasInformedBy relation from an informed prov:Activity to the prov:Activity that informed it. For example, :you_jumping_off_bridge prov:wasInformedBy :everyone_else_jumping_off_bridge; prov:qualifiedCommunication [ a prov:Communication; prov:activity :everyone_else_jumping_off_bridge; :foo :bar ].
    Contribute: URIRef  # Contribute
    Contributor: URIRef  # Contributor
    Copyright: URIRef  # Copyright
    Create: URIRef  # Create
    Creator: URIRef  # Creator
    Delegation: URIRef  # An instance of prov:Delegation provides additional descriptions about the binary prov:actedOnBehalfOf relation from a performing prov:Agent to some prov:Agent for whom it was performed. For example, :mixing prov:wasAssociatedWith :toddler . :toddler prov:actedOnBehalfOf :mother; prov:qualifiedDelegation [ a prov:Delegation; prov:entity :mother; :foo :bar ].
    Derivation: URIRef  # The more specific forms of prov:Derivation (i.e., prov:Revision, prov:Quotation, prov:PrimarySource) should be asserted if they apply.
    Dictionary: URIRef  # This concept allows for the provenance of the dictionary, but also of its constituents to be expressed. Such a notion of dictionary corresponds to a wide variety of concrete data structures, such as a maps or associative arrays.
    DirectQueryService: URIRef  # Type for a generic provenance query service. Mainly for use in RDF provenance query service descriptions, to facilitate discovery in linked data environments.
    EmptyDictionary: URIRef  # Empty Dictionary
    End: URIRef  # An instance of prov:End provides additional descriptions about the binary prov:wasEndedBy relation from some ended prov:Activity to an prov:Entity that ended it. For example, :ball_game prov:wasEndedBy :buzzer; prov:qualifiedEnd [ a prov:End; prov:entity :buzzer; :foo :bar; prov:atTime '2012-03-09T08:05:08-05:00'^^xsd:dateTime ].
    Entity: URIRef  # Entity
    EntityInfluence: URIRef  # It is not recommended that the type EntityInfluence be asserted without also asserting one of its more specific subclasses.
    Generation: URIRef  # An instance of prov:Generation provides additional descriptions about the binary prov:wasGeneratedBy relation from a generated prov:Entity to the prov:Activity that generated it. For example, :cake prov:wasGeneratedBy :baking; prov:qualifiedGeneration [ a prov:Generation; prov:activity :baking; :foo :bar ].
    Influence: URIRef  # Because prov:Influence is a broad relation, its most specific subclasses (e.g. prov:Communication, prov:Delegation, prov:End, prov:Revision, etc.) should be used when applicable.
    Insertion: URIRef  # Insertion
    InstantaneousEvent: URIRef  # An instantaneous event, or event for short, happens in the world and marks a change in the world, in its activities and in its entities. The term 'event' is commonly used in process algebra with a similar meaning. Events represent communications or interactions; they are assumed to be atomic and instantaneous.
    Invalidation: URIRef  # An instance of prov:Invalidation provides additional descriptions about the binary prov:wasInvalidatedBy relation from an invalidated prov:Entity to the prov:Activity that invalidated it. For example, :uncracked_egg prov:wasInvalidatedBy :baking; prov:qualifiedInvalidation [ a prov:Invalidation; prov:activity :baking; :foo :bar ].
    KeyEntityPair: URIRef  # Key-Entity Pair
    Location: URIRef  # Location
    Modify: URIRef  # Modify
    Organization: URIRef  # Organization
    Person: URIRef  # Person
    Plan: URIRef  # There exist no prescriptive requirement on the nature of plans, their representation, the actions or steps they consist of, or their intended goals. Since plans may evolve over time, it may become necessary to track their provenance, so plans themselves are entities. Representing the plan explicitly in the provenance can be useful for various tasks: for example, to validate the execution as represented in the provenance record, to manage expectation failures, or to provide explanations.
    PrimarySource: URIRef  # An instance of prov:PrimarySource provides additional descriptions about the binary prov:hadPrimarySource relation from some secondary prov:Entity to an earlier, primary prov:Entity. For example, :blog prov:hadPrimarySource :newsArticle; prov:qualifiedPrimarySource [ a prov:PrimarySource; prov:entity :newsArticle; :foo :bar ] .
    Publish: URIRef  # Publish
    Publisher: URIRef  # Publisher
    Quotation: URIRef  # An instance of prov:Quotation provides additional descriptions about the binary prov:wasQuotedFrom relation from some taken prov:Entity from an earlier, larger prov:Entity. For example, :here_is_looking_at_you_kid prov:wasQuotedFrom :casablanca_script; prov:qualifiedQuotation [ a prov:Quotation; prov:entity :casablanca_script; :foo :bar ].
    Removal: URIRef  # Removal
    Replace: URIRef  # Replace
    Revision: URIRef  # An instance of prov:Revision provides additional descriptions about the binary prov:wasRevisionOf relation from some newer prov:Entity to an earlier prov:Entity. For example, :draft_2 prov:wasRevisionOf :draft_1; prov:qualifiedRevision [ a prov:Revision; prov:entity :draft_1; :foo :bar ].
    RightsAssignment: URIRef  # RightsAssignment
    RightsHolder: URIRef  # RightsHolder
    Role: URIRef  # Role
    ServiceDescription: URIRef  # Type for a generic provenance query service. Mainly for use in RDF provenance query service descriptions, to facilitate discovery in linked data environments.
    SoftwareAgent: URIRef  # SoftwareAgent
    Start: URIRef  # An instance of prov:Start provides additional descriptions about the binary prov:wasStartedBy relation from some started prov:Activity to an prov:Entity that started it. For example, :foot_race prov:wasStartedBy :bang; prov:qualifiedStart [ a prov:Start; prov:entity :bang; :foo :bar; prov:atTime '2012-03-09T08:05:08-05:00'^^xsd:dateTime ] .
    Submit: URIRef  # Submit
    Usage: URIRef  # An instance of prov:Usage provides additional descriptions about the binary prov:used relation from some prov:Activity to an prov:Entity that it used. For example, :keynote prov:used :podium; prov:qualifiedUsage [ a prov:Usage; prov:entity :podium; :foo :bar ].

    # http://www.w3.org/2002/07/owl#DatatypeProperty
    atTime: URIRef  # The time at which an InstantaneousEvent occurred, in the form of xsd:dateTime.
    endedAtTime: (
        URIRef  # The time at which an activity ended. See also prov:startedAtTime.
    )
    generatedAtTime: URIRef  # The time at which an entity was completely created and is available for use.
    invalidatedAtTime: (
        URIRef  # The time at which an entity was invalidated (i.e., no longer usable).
    )
    provenanceUriTemplate: URIRef  # Relates a provenance service to a URI template string for constructing provenance-URIs.
    removedKey: URIRef  # removedKey
    startedAtTime: (
        URIRef  # The time at which an activity started. See also prov:endedAtTime.
    )
    value: URIRef  # value

    # http://www.w3.org/2002/07/owl#FunctionalProperty
    pairEntity: URIRef  # pairKey
    pairKey: URIRef  # pairKey

    # http://www.w3.org/2002/07/owl#NamedIndividual
    EmptyCollection: URIRef  # EmptyCollection

    # http://www.w3.org/2002/07/owl#ObjectProperty
    actedOnBehalfOf: URIRef  # An object property to express the accountability of an agent towards another agent. The subordinate agent acted on behalf of the responsible agent in an actual activity.
    activity: URIRef  # activity
    agent: URIRef  # agent
    alternateOf: URIRef  # alternateOf
    asInBundle: URIRef  # prov:asInBundle is used to specify which bundle the general entity of a prov:mentionOf property is described.  When :x prov:mentionOf :y and :y is described in Bundle :b, the triple :x prov:asInBundle :b is also asserted to cite the Bundle in which :y was described.
    atLocation: URIRef  # The Location of any resource.
    derivedByInsertionFrom: URIRef  # derivedByInsertionFrom
    derivedByRemovalFrom: URIRef  # derivedByRemovalFrom
    describesService: URIRef  # relates a generic provenance query service resource (type prov:ServiceDescription) to a specific query service description (e.g. a prov:DirectQueryService or a sd:Service).
    dictionary: URIRef  # dictionary
    entity: URIRef  # entity
    generated: URIRef  # generated
    hadActivity: URIRef  # The _optional_ Activity of an Influence, which used, generated, invalidated, or was the responsibility of some Entity. This property is _not_ used by ActivityInfluence (use prov:activity instead).
    hadDictionaryMember: URIRef  # hadDictionaryMember
    hadGeneration: (
        URIRef  # The _optional_ Generation involved in an Entity's Derivation.
    )
    hadMember: URIRef  # hadMember
    hadPlan: URIRef  # The _optional_ Plan adopted by an Agent in Association with some Activity. Plan specifications are out of the scope of this specification.
    hadPrimarySource: URIRef  # hadPrimarySource
    hadRole: URIRef  # This property has multiple RDFS domains to suit multiple OWL Profiles. See <a href="#owl-profile">PROV-O OWL Profile</a>.
    hadUsage: URIRef  # The _optional_ Usage involved in an Entity's Derivation.
    has_anchor: (
        URIRef  # Indicates anchor URI for a potentially dynamic resource instance.
    )
    has_provenance: URIRef  # Indicates a provenance-URI for a resource; the resource identified by this property presents a provenance record about its subject or anchor resource.
    has_query_service: URIRef  # Indicates a provenance query service that can access provenance related to its subject or anchor resource.
    influenced: URIRef  # influenced
    influencer: URIRef  # Subproperties of prov:influencer are used to cite the object of an unqualified PROV-O triple whose predicate is a subproperty of prov:wasInfluencedBy (e.g. prov:used, prov:wasGeneratedBy). prov:influencer is used much like rdf:object is used.
    insertedKeyEntityPair: URIRef  # insertedKeyEntityPair
    invalidated: URIRef  # invalidated
    mentionOf: URIRef  # prov:mentionOf is used to specialize an entity as described in another bundle. It is to be used in conjunction with prov:asInBundle.  prov:asInBundle is used to cite the Bundle in which the generalization was mentioned.
    pingback: URIRef  # Relates a resource to a provenance pingback service that may receive additional provenance links about the resource.
    qualifiedAssociation: URIRef  # If this Activity prov:wasAssociatedWith Agent :ag, then it can qualify the Association using prov:qualifiedAssociation [ a prov:Association;  prov:agent :ag; :foo :bar ].
    qualifiedAttribution: URIRef  # If this Entity prov:wasAttributedTo Agent :ag, then it can qualify how it was influenced using prov:qualifiedAttribution [ a prov:Attribution;  prov:agent :ag; :foo :bar ].
    qualifiedCommunication: URIRef  # If this Activity prov:wasInformedBy Activity :a, then it can qualify how it was influenced using prov:qualifiedCommunication [ a prov:Communication;  prov:activity :a; :foo :bar ].
    qualifiedDelegation: URIRef  # If this Agent prov:actedOnBehalfOf Agent :ag, then it can qualify how with prov:qualifiedResponsibility [ a prov:Responsibility;  prov:agent :ag; :foo :bar ].
    qualifiedDerivation: URIRef  # If this Entity prov:wasDerivedFrom Entity :e, then it can qualify how it was derived using prov:qualifiedDerivation [ a prov:Derivation;  prov:entity :e; :foo :bar ].
    qualifiedEnd: URIRef  # If this Activity prov:wasEndedBy Entity :e1, then it can qualify how it was ended using prov:qualifiedEnd [ a prov:End;  prov:entity :e1; :foo :bar ].
    qualifiedGeneration: URIRef  # If this Activity prov:generated Entity :e, then it can qualify how it performed the Generation using prov:qualifiedGeneration [ a prov:Generation;  prov:entity :e; :foo :bar ].
    qualifiedInfluence: URIRef  # Because prov:qualifiedInfluence is a broad relation, the more specific relations (qualifiedCommunication, qualifiedDelegation, qualifiedEnd, etc.) should be used when applicable.
    qualifiedInsertion: URIRef  # qualifiedInsertion
    qualifiedInvalidation: URIRef  # If this Entity prov:wasInvalidatedBy Activity :a, then it can qualify how it was invalidated using prov:qualifiedInvalidation [ a prov:Invalidation;  prov:activity :a; :foo :bar ].
    qualifiedPrimarySource: URIRef  # If this Entity prov:hadPrimarySource Entity :e, then it can qualify how using prov:qualifiedPrimarySource [ a prov:PrimarySource; prov:entity :e; :foo :bar ].
    qualifiedQuotation: URIRef  # If this Entity prov:wasQuotedFrom Entity :e, then it can qualify how using prov:qualifiedQuotation [ a prov:Quotation;  prov:entity :e; :foo :bar ].
    qualifiedRemoval: URIRef  # qualifiedRemoval
    qualifiedRevision: URIRef  # If this Entity prov:wasRevisionOf Entity :e, then it can qualify how it was revised using prov:qualifiedRevision [ a prov:Revision;  prov:entity :e; :foo :bar ].
    qualifiedStart: URIRef  # If this Activity prov:wasStartedBy Entity :e1, then it can qualify how it was started using prov:qualifiedStart [ a prov:Start;  prov:entity :e1; :foo :bar ].
    qualifiedUsage: URIRef  # If this Activity prov:used Entity :e, then it can qualify how it used it using prov:qualifiedUsage [ a prov:Usage; prov:entity :e; :foo :bar ].
    used: URIRef  # A prov:Entity that was used by this prov:Activity. For example, :baking prov:used :spoon, :egg, :oven .
    wasAssociatedWith: URIRef  # An prov:Agent that had some (unspecified) responsibility for the occurrence of this prov:Activity.
    wasAttributedTo: URIRef  # Attribution is the ascribing of an entity to an agent.
    wasDerivedFrom: URIRef  # The more specific subproperties of prov:wasDerivedFrom (i.e., prov:wasQuotedFrom, prov:wasRevisionOf, prov:hadPrimarySource) should be used when applicable.
    wasEndedBy: URIRef  # End is when an activity is deemed to have ended. An end may refer to an entity, known as trigger, that terminated the activity.
    wasGeneratedBy: URIRef  # wasGeneratedBy
    wasInfluencedBy: URIRef  # This property has multiple RDFS domains to suit multiple OWL Profiles. See <a href="#owl-profile">PROV-O OWL Profile</a>.
    wasInformedBy: URIRef  # An activity a2 is dependent on or informed by another activity a1, by way of some unspecified entity that is generated by a1 and used by a2.
    wasInvalidatedBy: URIRef  # wasInvalidatedBy
    wasQuotedFrom: URIRef  # An entity is derived from an original entity by copying, or 'quoting', some or all of it.
    wasStartedBy: URIRef  # Start is when an activity is deemed to have started. A start may refer to an entity, known as trigger, that initiated the activity.

    _NS = Namespace("http://www.w3.org/ns/prov#")



from rdflib import RDF, Graph, Namespace

EX = Namespace("http://example.org/")


def test_items():
    g = Graph().parse(
        data="""
        @prefix : <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .

        <> :value (
            <http://example.org/thing1>
            <http://example.org/thing2>
        ).

        <> :value (
            <http://example.org/thing3>
        ).

        <> :value ().

        <> :value (
            <http://example.org/thing4>
            <http://example.org/thing5>
            <http://example.org/thing6>
        ).
        """,
        format="turtle",
    )

    values = {tuple(g.items(v)) for v in g.objects(None, RDF.value)}
    assert values == {
        (EX.thing1, EX.thing2),
        (EX.thing3,),
        (),
        (EX.thing4, EX.thing5, EX.thing6),
    }


def test_recursive_list_detection():
    g = Graph().parse(
        data="""
        @prefix : <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .

        <> :value _:a .
        _:a :first "turtles"; :rest _:a .

        <> :value [ :first "turtles"; :rest _:b ] .
        _:b :first "all the way down"; :rest _:b .

        <> :value [ :first "turtles"; :rest _:c ] .
        _:c :first "all the way down"; :rest _:a .

        """,
        format="turtle",
    )

    for v in g.objects(None, RDF.value):
        try:
            list(g.items(v))
        except ValueError as e:  # noqa: F841
            pass
        else:
            assert False, "Expected detection of recursive rdf:rest reference"



from io import StringIO

from rdflib import logger, plugin
from rdflib.graph import ConjunctiveGraph, Graph, ReadOnlyGraphAggregate
from rdflib.namespace import RDF, RDFS
from rdflib.store import Store
from rdflib.term import URIRef

TEST_GRAPH_1N3 = """
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix : <http://test/> .
:foo a rdfs:Class.
:bar :d :c.
:a :d :c.
"""


TEST_GRAPH_2N3 = """
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix : <http://test/> .
@prefix log: <http://www.w3.org/2000/10/swap/log#>.
:foo a rdfs:Resource.
:bar rdfs:isDefinedBy [ a log:Formula ].
:a :d :e.
"""

TEST_GRAPH_3N3 = """
@prefix rdf:  <http://www.w3.org/1999/02/22-rdf-syntax-ns#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix log: <http://www.w3.org/2000/10/swap/log#>.
@prefix : <http://test/> .
<> a log:N3Document.
"""

SPARQL_Q = """
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT *
FROM NAMED <http://example.com/graph1>
FROM NAMED <http://example.com/graph2>
FROM NAMED <http://example.com/graph3>
FROM <http://www.w3.org/2000/01/rdf-schema#>

WHERE {?sub ?pred rdfs:Class }"""

SPARQL_Q2 = """
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT ?class
WHERE { GRAPH ?graph { ?member a ?class } }"""

SPARQL_Q3 = """
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX log: <http://www.w3.org/2000/10/swap/log#>
SELECT ?n3Doc
WHERE {?n3Doc a log:N3Document }"""


def test_aggregate_raw():
    mem_store = plugin.get("Memory", Store)()
    graph1 = Graph(mem_store)
    graph2 = Graph(mem_store)
    graph3 = Graph(mem_store)

    for n3_str, graph in [
        (TEST_GRAPH_1N3, graph1),
        (TEST_GRAPH_2N3, graph2),
        (TEST_GRAPH_3N3, graph3),
    ]:
        graph.parse(StringIO(n3_str), format="n3")

    g = ReadOnlyGraphAggregate([graph1, graph2, graph3])

    # Test triples
    assert len(list(g.triples((None, RDF.type, None)))) == 4
    assert len(list(g.triples((URIRef("http://test/bar"), None, None)))) == 2
    assert len(list(g.triples((None, URIRef("http://test/d"), None)))) == 3

    # Test __len__
    assert len(g) == 8

    # assert context iteration
    for g in g.contexts():
        assert isinstance(g, Graph)

    # Test __contains__
    assert (URIRef("http://test/foo"), RDF.type, RDFS.Resource) in g

    bar_predicates = [URIRef("http://test/d"), RDFS.isDefinedBy]
    assert (
        len(list(g.triples_choices((URIRef("http://test/bar"), bar_predicates, None))))
        == 2
    )


def test_aggregate2():
    mem_store = plugin.get("Memory", Store)()
    graph1 = Graph(mem_store, URIRef("http://example.com/graph1"))
    graph2 = Graph(mem_store, URIRef("http://example.com/graph2"))
    graph3 = Graph(mem_store, URIRef("http://example.com/graph3"))

    for n3_str, graph in [
        (TEST_GRAPH_1N3, graph1),
        (TEST_GRAPH_2N3, graph2),
        (TEST_GRAPH_3N3, graph3),
    ]:
        graph.parse(StringIO(n3_str), format="n3")

    graph4 = Graph(mem_store, RDFS)
    graph4.parse(data=TEST_GRAPH_1N3, format="n3")
    g = ConjunctiveGraph(mem_store)
    assert g is not None
    assert len(list(g.quads((None, None, None, None)))) == 11
    assert len(list(g.contexts())) == 4
    logger.debug(list(g.contexts()))
    assert (
        len(list(g.quads((None, None, None, URIRef("http://example.com/graph2"))))) == 4
    )
    assert (
        len(
            list(
                g.quads(
                    (None, None, None, URIRef("http://www.w3.org/2000/01/rdf-schema#"))
                )
            )
        )
        == 6
    )



from rdflib import Graph, Literal, Namespace, URIRef
from rdflib.extras.infixowl import Class, Ontology, Property, min, only, some

CPR = Namespace("http://purl.org/cpr/0.75#")
INF = Namespace("http://www.loa-cnr.it/ontologies/InformationObjects.owl#")
EDNS = Namespace("http://www.loa-cnr.it/ontologies/ExtendedDnS.owl#")
DOLCE = Namespace("http://www.loa-cnr.it/ontologies/DOLCE-Lite.owl#")
REL = Namespace("http://www.geneontology.org/owl#")
GALEN = Namespace("http://www.co-ode.org/ontologies/galen#")
TIME = Namespace("http://www.w3.org/2006/time#")
CYC = Namespace("http://sw.cyc.com/2006/07/27/cyc/")


def infixowl_example():
    g = Graph()
    g.bind("cpr", CPR, override=False)
    g.bind("ro", REL, override=False)
    g.bind("inf", INF, override=False)
    g.bind("edns", EDNS, override=False)
    g.bind("dol", DOLCE, override=False)
    g.bind("time", TIME, override=False)
    g.bind("galen", GALEN, override=False)

    Class.factoryGraph = g
    Property.factoryGraph = g
    Ontology.factoryGraph = g

    cprOntology = Ontology(URIRef("http://purl.org/cpr/owl"))  # noqa: N806
    cprOntology.imports = [
        URIRef("http://obo.sourceforge.net/relationship/relationship.owl"),
        URIRef(DOLCE),
        URIRef(EDNS),
        URIRef("http://www.w3.org/2006/time#"),
    ]
    cprOntology.comment = [
        Literal(
            """This OWL ontology was generated by Fuxi 0.85b.dev-r107
               (with newly added Infix OWL syntax library).  It imports the
               OBO relationship ontology, DOLCE, and OWL time. It formally
               defines a focused, core set of archetypes [Jung, C.]
               replicated in various patient record terminology. This core is
               defined in RDF and follows the normalization principles
               of "rigorous formal ontologies" [Rector, A.]."""
        )
    ]
    cprOntology.setVersion(Literal("0.75"))

    # Relations
    # represented-by
    representationOf = Property(  # noqa: N806
        CPR["representation-of"],
        inverseOf=Property(CPR["represented-by"]),
        comment=[
            Literal(
                """Patient records stand in the cpr:representation-of relation
                   with patients"""
            )
        ],
    )
    representedBy = Property(  # noqa: F841, N806
        CPR["represented-by"], inverseOf=representationOf
    )
    # description-of
    descrOf = Property(  # noqa: N806
        CPR["description-of"],
        comment=[
            Literal(
                """Clinical descriptions stand in the cpr:description-of
                   relation with various clinical phenomenon"""
            )
        ],
        domain=[Class(CPR["clinical-description"])],
    )
    # cpr:interpreted-by
    interpretedBy = Property(  # noqa: F841, N806
        CPR["interpreted-by"],
        comment=[
            Literal(
                """Signs and symptoms are interpreted by rational physical
                   objects (people)"""
            )
        ],
        domain=[Class(CPR["medical-sign"]) | Class(CPR["symptom"])],
        range=[Class(CPR.person)],
    )
    # cpr:realized-by
    realizedBy = Property(  # noqa: N806
        CPR["realized-by"],
        comment=[
            Literal(
                """The epistemological relation in which screening acts and
                   the problems they realize stand to each other"""
            )
        ],
        inverseOf=Property(CPR["realizes"]),
        domain=[Class(CPR["medical-problem"])],
        range=[Class(CPR["screening-act"])],
    )
    # cpr:realizes
    realizes = Property(CPR["realizes"], inverseOf=realizedBy)  # noqa: F841

    # Classes
    # cpr:person
    person = Class(CPR.person)
    person.comment = [
        Literal(
            """A class which directly corresponds with the “Person” class in
               both GALEN and Cyc"""
        )
    ]
    person.subClassOf = [Class(EDNS["rational-physical-object"])]
    person.equivalentClass = [Class(GALEN.Person), Class(CYC.Person)]

    # cpr:patient
    patient = Class(CPR.patient)
    patient.comment = [
        Literal(
            """A class which directly corresponds with the “Patient”
               and “MedicalPatient” classes in GALEN / Cyc"""
        )
    ]
    # patient.equivalentClass = [Class(GALEN.Patient),Class(CYC.MedicalPatient)]
    patient.subClassOf = [CPR["represented-by"] @ some @ Class(CPR["patient-record"])]
    person += patient

    # cpr:clinician
    clinician = Class(CPR.person)
    clinician.comment = [
        Literal(
            """A person who plays the clinician role (typically Nurse,
               Physician / Doctor, etc.)"""
        )
    ]
    person += clinician

    # bytes
    bytes = Class(CPR.bytes)
    bytes.comment = [
        Literal(
            """The collection of physical objects which constitute a stream of
               bytes in memory, disk, etc."""
        )
    ]
    bytes.subClassOf = [DOLCE["non-agentive-physical-object"]]

    # cpr:patient-record
    patientRecord = Class(CPR["patient-record"])  # noqa: N806
    patientRecord.comment = [
        Literal(
            """a class (a representational artifact [REFTERM]) depicting
               relevant clinical information about a specific patient and is
               primarily comprised of one or more
               cpr:clinical-descriptions."""
        )
    ]
    patientRecord.seeAlso = [URIRef("")]
    patientRecord.subClassOf = [
        bytes,
        # Class(CYC.InformationBearingThing),
        CPR["representation-of"] @ only @ patient,
        REL.OBO_REL_has_proper_part @ some @ Class(CPR["clinical-description"]),
    ]

    # cpr:medical-problem
    problem = Class(
        CPR["medical-problem"],
        subClassOf=[
            Class(DOLCE.quality),
            realizedBy @ only @ Class(CPR["screening-act"]),
        ],
    )
    problem.comment = [
        Literal(
            """.. problems that clearly require the intervention of a health
                  care professional. These include acute problems requiring
                  hospitalization and chronic problems requiring long-term
                  management."""
        )
    ]

    # cpr:clinical-description
    clinDescr = Class(CPR["clinical-description"])  # noqa: N806
    clinDescr.disjointWith = [CPR["patient-record"]]
    clinDescr.comment = [
        Literal(
            """A class which corresponds (at least syntactically) with the HL7
               RIM Act Class, insofar as its members consist of clinical
               recordings (representational artifacts) of natural phenomena
               of clinical significance"""
        )
    ]
    clinDescr.subClassOf = [
        bytes,
        # Class(CYC.InformationBearingThing),
        DOLCE["has-quality"] @ some @ Class(TIME.TemporalEntity),
        descrOf @ min @ Literal(1),
    ]

    # cpr:medical-sign
    sign = Class(
        CPR["medical-sign"],
        subClassOf=[
            problem,
            Property(CPR["interpreted-by"]) @ only @ clinician,
            Property(CPR["interpreted-by"]) @ some @ clinician,
        ],
        disjointWith=[CPR.symptom],
    )
    sign.comment = [
        Literal(
            """A cpr:medical-problem which are specifically interpreted by a
               clinician. As such, this class is informally defined as an
               objective indication of a quality typically detected by a
               physician during a physical examination of a patient."""
        )
    ]

    symptom = Class(
        CPR["symptom"],
        subClassOf=[
            problem,
            Property(CPR["interpreted-by"]) @ only @ patient,
            Property(CPR["interpreted-by"]) @ some @ patient,
        ],
        disjointWith=[sign],
    )
    symptom.comment = [
        Literal(
            """(Medicine) any sensation or change in bodily function that is
               experienced by a patient and is associated with a particular
               disease."""
        )
    ]

    # clinical-act heriarchy
    clinicalAct = Class(  # noqa: N806
        CPR["clinical-act"], subClassOf=[Class(EDNS.activity)]
    )

    therapy = Class(CPR["therapeutic-act"], subClassOf=[clinicalAct])
    therapy += Class(CPR["physical-therapy"], disjointWith=[CPR["medical-therapy"]])
    therapy += Class(
        CPR["psychological-therapy"],
        disjointWith=[CPR["medical-therapy"], CPR["physical-therapy"]],
    )

    medicalTherapy = Class(  # noqa: N806
        CPR["medical-therapy"],
        disjointWith=[CPR["physical-therapy"], CPR["psychological-therapy"]],
    )
    therapy += medicalTherapy
    medicalTherapy += Class(CPR["substance-administration"])  # noqa: N806

    diagnosticAct = Class(CPR["diagnostic-act"], subClassOf=[clinicalAct])  # noqa: N806
    diagnosticAct.disjointWith = [CPR["therapeutic-act"]]

    screeningAct = Class(CPR["screening-act"])  # noqa: N806
    screeningAct += Class(CPR["laboratory-test"])  # noqa: N806

    diagnosticAct += screeningAct  # noqa: N806

    screeningAct += Class(  # noqa: N806
        CPR["medical-history-screening-act"],
        disjointWith=[CPR["clinical-examination"], CPR["laboratory-test"]],
    )

    screeningAct += Class(  # noqa: N806
        CPR["clinical-examination"],
        disjointWith=[CPR["laboratory-test"], CPR["medical-history-screening-act"]],
    )

    device = Class(  # noqa: F841
        CPR["medical-device"], subClassOf=[Class(GALEN.Device)]
    )

    print(g.serialize(format="turtle"))


if __name__ == "__main__":
    infixowl_example()



import os

from rdflib import Graph, URIRef
from rdflib.namespace import FOAF
from rdflib.plugins.sparql import prepareQuery, prepareUpdate


def test_prepare_update():
    q = prepareUpdate(
        """\
PREFIX dc: <http://purl.org/dc/elements/1.1/>
INSERT DATA
{ <http://example/book3> dc:title "A new book" ;
                         dc:creator "A.N.Other" .
 } ;
""",
        initNs={},
    )

    g = Graph()
    g.update(q, initBindings={})
    assert len(g) == 2


def test_prepare_query():
    q = prepareQuery(
        "SELECT ?name WHERE { ?person foaf:knows/foaf:name ?name . }",
        initNs={"foaf": FOAF},
    )

    g = Graph()
    g.parse(
        location=os.path.join(
            os.path.dirname(__file__), "..", "..", "examples", "foaf.n3"
        ),
        format="n3",
    )

    tim = URIRef("http://www.w3.org/People/Berners-Lee/card#i")

    assert len(list(g.query(q, initBindings={"person": tim}))) == 50



import os
import re
import subprocess
import sys
from tempfile import mkstemp
from unittest.mock import mock_open, patch, sentinel

from rdflib.tools import csv2rdf
from test.data import TEST_DATA_DIR

REALESTATE_FILE_PATH = os.path.join(TEST_DATA_DIR, "csv", "realestate.csv")


class TestCSV2RDF:
    def test_csv2rdf_cli(self):
        completed = subprocess.run(
            [
                sys.executable,
                "-m",
                "rdflib.tools.csv2rdf",
                str(REALESTATE_FILE_PATH),
            ],
            capture_output=True,
            text=True,
        )
        assert completed.returncode == 0
        assert "Converted 19 rows into 228 triples." in completed.stderr
        assert re.search(r"Took [\d\.]+ seconds\.", completed.stderr)
        assert "<http://example.org/instances/0>" in completed.stdout
        assert "<http://example.org/props/zip>" in completed.stdout

    def test_csv2rdf_cli_fileout(self):
        fh, fname = mkstemp()
        completed = subprocess.run(
            [
                sys.executable,
                "-m",
                "rdflib.tools.csv2rdf",
                "-o",
                fname,
                str(REALESTATE_FILE_PATH),
            ],
        )
        assert completed.returncode == 0
        with open(fname) as f:
            assert len(f.readlines()) == 228
        os.close(fh)
        os.remove(fname)

    @patch.object(csv2rdf.CSV2RDF, "convert", return_value=None)
    def test_csv2rdf_config_file_opened(self, config_mock):
        """Test that the config file is read when specified."""
        # Pretend there is a file with the section we're looking for.
        # We don't care about the actual path, since it won't really be opened
        # but when we try to open it, we will get back the section header
        # so that the reader doesn't complain.
        config_file = sentinel.file_path
        open_mock = mock_open(read_data="[csv2rdf]")
        # Also pretend that we're passing the arguments from the command line
        cli_args = ["csv2rdf.py", "-f", config_file, str(REALESTATE_FILE_PATH)]
        with patch.object(csv2rdf.sys, "argv", cli_args):
            with patch("builtins.open", open_mock):
                csv2rdf.main()
                # Check that we've "opened" the right file
                open_mock.assert_called_once_with(config_file)



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class ORG(DefinedNamespace):
    """
    Core organization ontology

    Vocabulary for describing organizational structures, specializable to a broad variety of types of
    organization.

    Generated from: http://www.w3.org/ns/org#
    Date: 2020-05-26 14:20:02.908408

    """

    _fail = True

    # http://www.w3.org/1999/02/22-rdf-syntax-ns#Property
    basedAt: URIRef  # Indicates the site at which a person is based. We do not restrict the possibility that a person is based at multiple sites.
    changedBy: URIRef  # Indicates a change event which resulted in a change to this organization. Depending on the event the organization may or may not have continued to exist after the event. Inverse of `org:originalOrganization`.
    classification: URIRef  # Indicates a classification for this Organization within some classification scheme. Extension vocabularies may wish to specialize this property to have a range corresponding to a specific `skos:ConceptScheme`. This property is under discussion and may be revised or removed - in many cases organizations are best categorized by defining a sub-class hierarchy in an extension vocabulary.
    hasMember: URIRef  # Indicates a person who is a member of the subject Organization. Inverse of `org:memberOf`, see that property for further clarification. Provided for compatibility with `foaf:member`.
    hasMembership: URIRef  # Indicates a membership relationship that the Agent plays. Inverse of `org:member`.
    hasPost: URIRef  # Indicates a Post which exists within the Organization.
    hasPrimarySite: URIRef  # Indicates a primary site for the Organization, this is the default means by which an Organization can be contacted and is not necessarily the formal headquarters.
    hasRegisteredSite: URIRef  # Indicates the legally registered site for the organization, in many legal jurisdictions there is a requirement that FormalOrganizations such as Companies or Charities have such a primary designed site.
    hasSite: URIRef  # Indicates a site at which the Organization has some presence even if only indirect (e.g. virtual office or a professional service which is acting as the registered address for a company). Inverse of `org:siteOf`.
    hasSubOrganization: URIRef  # Represents hierarchical containment of Organizations or Organizational Units; indicates an organization which is a sub-part or child of this organization.  Inverse of `org:subOrganizationOf`.
    hasUnit: URIRef  # Indicates a unit which is part of this Organization, e.g. a Department within a larger FormalOrganization. Inverse of `org:unitOf`.
    headOf: URIRef  # Indicates that a person is the leader or formal head of the Organization. This will normally mean that they are the root of the `org:reportsTo` (acyclic) graph, though an organization may have more than one head.
    heldBy: URIRef  # Indicates an Agent which holds a Post.
    holds: URIRef  # Indicates a Post held by some Agent.
    identifier: URIRef  # Gives an identifier, such as a company registration number, that can be used to used to uniquely identify the organization. Many different national and international identier schemes are available. The org ontology is neutral to which schemes are used. The particular identifier scheme should be indicated by the datatype of the identifier value.  Using datatypes to distinguish the notation scheme used is consistent with recommended best practice for `skos:notation` of which this property is a specialization.
    linkedTo: URIRef  # Indicates an arbitrary relationship between two organizations. Specializations of this can be used to, for example, denote funding or supply chain relationships.
    location: URIRef  # Gives a location description for a person within the organization, for example a _Mail Stop_ for internal posting purposes.
    member: URIRef  # Indicates the Person (or other Agent including Organization) involved in the Membership relationship. Inverse of `org:hasMembership`
    memberDuring: URIRef  # Optional property to indicate the interval for which the membership is/was valid.
    memberOf: URIRef  # Indicates that a person is a member of the Organization with no indication of the nature of that membership or the role played. Note that the choice of property name is not meant to limit the property to only formal membership arrangements, it is also intended to cover related concepts such as affilliation or other involvement in the organization. Extensions can specialize this relationship to indicate particular roles within the organization or more nuanced relationships to the organization. Has an optional inverse, `org:hasmember`.
    organization: URIRef  # Indicates Organization in which the Agent is a member.
    originalOrganization: URIRef  # Indicates one or more organizations that existed before the change event. Depending on the event they may or may not have continued to exist after the event. Inverse of `org:changedBy`.
    postIn: URIRef  # Indicates the Organization in which the Post exists.
    purpose: URIRef  # Indicates the purpose of this Organization. There can be many purposes at different levels of abstraction but the nature of an organization is to have a reason for existence and this property is a means to document that reason. An Organization may have multiple purposes. It is recommended that the purpose be denoted by a controlled term or code list, ideally a `skos:Concept`. However, the range is left open to allow for other types of descriptive schemes. It is expected that specializations or application profiles of this vocabulary will constrain the range of the purpose. Alternative names: _remit_ _responsibility_ (esp. if applied to OrganizationalUnits such as Government Departments).
    remuneration: URIRef  # Indicates a salary or other reward associated with the role. Typically this will be denoted using an existing representation scheme such as `gr:PriceSpecification` but the range is left open to allow applications to specialize it (e.g. to remunerationInGBP).
    reportsTo: URIRef  # Indicates a reporting relationship as might be depicted on an organizational chart. The precise semantics of the reporting relationship will vary by organization but is intended to encompass both direct supervisory relationships (e.g. carrying objective and salary setting authority) and more general reporting or accountability relationships (e.g. so called _dotted line_ reporting).
    resultedFrom: URIRef  # Indicates an event which resulted in this organization. Inverse of `org:resultingOrganization`.
    resultingOrganization: URIRef  # Indicates an organization which was created or changed as a result of the event. Inverse of `org:resultedFrom`.
    role: URIRef  # Indicates the Role that the Agent plays in a Membership relationship with an Organization.
    roleProperty: URIRef  # This is a metalevel property which is used to annotate an `org:Role` instance with a sub-property of `org:memberOf` that can be used to directly indicate the role for easy of query. The intended semantics is a Membership relation involving the Role implies the existence of a direct property relationship through an inference rule of the form:  `{ [] org:member ?p; org:organization ?o; org:role [org:roleProperty ?r] } -> {?p ?r ?o}`.
    siteAddress: URIRef  # Indicates an address for the site in a suitable encoding. Use of vCard (using the http://www.w3.org/TR/vcard-rdf/ vocabulary) is encouraged but the range is left open to allow other encodings to be used. The address may include email, telephone, and geo-location information and is not restricted to a physical address.
    siteOf: URIRef  # Indicates an Organization which has some presence at the given site. This is the inverse of `org:hasSite`.
    subOrganizationOf: URIRef  # Represents hierarchical containment of Organizations or OrganizationalUnits; indicates an Organization which contains this Organization. Inverse of `org:hasSubOrganization`.
    transitiveSubOrganizationOf: URIRef  # The transitive closure of subOrganizationOf, giving a representation of all organizations that contain this one. Note that technically this is a super property of the transitive closure so it could contain additional assertions but such usage is discouraged.
    unitOf: URIRef  # Indicates an Organization of which this Unit is a part, e.g. a Department within a larger FormalOrganization. This is the inverse of `org:hasUnit`.

    # http://www.w3.org/2000/01/rdf-schema#Class
    ChangeEvent: URIRef  # Represents an event which resulted in a major change to an organization such as a merger or complete restructuring. It is intended for situations where the resulting organization is sufficient distinct from the original organizations that it has a distinct identity and distinct URI. Extension vocabularies should define sub-classes of this to denote particular categories of event. The instant or interval at which the event occurred should be given by `prov:startAtTime` and `prov:endedAtTime`, a description should be given by `dct:description`.
    FormalOrganization: URIRef  # An Organization which is recognized in the world at large, in particular in legal jurisdictions, with associated rights and responsibilities. Examples include a Corporation, Charity, Government or Church. Note that this is a super class of `gr:BusinessEntity` and it is recommended to use the GoodRelations vocabulary to denote Business classifications such as DUNS or NAICS.
    Membership: URIRef  # Indicates the nature of an Agent's membership of an organization. Represents an n-ary relation between an Agent, an Organization and a Role. It is possible to directly indicate membership, independent of the specific Role, through use of the `org:memberOf` property.
    Organization: URIRef  # Represents a collection of people organized together into a community or other social, commercial or political structure. The group has some common purpose or reason for existence which goes beyond the set of people belonging to it and can act as an Agent. Organizations are often decomposable into hierarchical structures.  It is recommended that SKOS lexical labels should be used to label the Organization. In particular `skos:prefLabel` for the primary (possibly legally recognized name), `skos:altLabel` for alternative names (trading names, colloquial names) and `skos:notation` to denote a code from a code list. Alternative names: _Collective_ _Body_ _Org_ _Group_
    OrganizationalCollaboration: URIRef  # A collaboration between two or more Organizations such as a project. It meets the criteria for being an Organization in that it has an identity and defining purpose independent of its particular members but is neither a formally recognized legal entity nor a sub-unit within some larger organization. Might typically have a shorter lifetime than the Organizations within it, but not necessarily. All members are `org:Organization`s rather than individuals and those Organizations can play particular roles within the venture. Alternative names: _Project_ _Venture_  _Endeavour_ _Consortium_ _Endeavour_
    OrganizationalUnit: URIRef  # An Organization such as a University Support Unit which is part of some larger FormalOrganization and only has full recognition within the context of that FormalOrganization, it is not a Legal Entity in its own right. Units can be large and complex containing other Units and even FormalOrganizations. Alternative names: _OU_ _Unit_ _Department_
    Post: URIRef  # A Post represents some position within an organization that exists independently of the person or persons filling it. Posts may be used to represent situations where a person is a member of an organization ex officio (for example the Secretary of State for Scotland is part of UK Cabinet by virtue of being Secretary of State for Scotland, not as an individual person). A post can be held by multiple people and hence can be treated as a organization in its own right.
    Role: URIRef  # Denotes a role that a Person or other Agent can take in an organization. Instances of this class describe the abstract role; to denote a specific instance of a person playing that role in a specific organization use an instance of `org:Membership`. It is common for roles to be arranged in some taxonomic structure and we use SKOS to represent that. The normal SKOS lexical properties should be used when labelling the Role. Additional descriptive properties for the Role, such as a Salary band, may be added by extension vocabularies.
    Site: URIRef  # An office or other premise at which the organization is located. Many organizations are spread across multiple sites and many sites will host multiple locations. In most cases a Site will be a physical location. However, we don't exclude the possibility of non-physical sites such as a virtual office with an associated post box and phone reception service. Extensions may provide subclasses to denote particular types of site.

    # http://www.w3.org/ns/org#Role
    Head: URIRef  # head

    _NS = Namespace("http://www.w3.org/ns/org#")



from __future__ import annotations

import os
from contextlib import ExitStack
from dataclasses import dataclass
from pathlib import PurePosixPath, PureWindowsPath
from typing import Any, List, Optional, Tuple, Type, Union

import pytest

from rdflib.graph import ConjunctiveGraph, Dataset, Graph
from rdflib.term import URIRef
from test.utils import (
    COLLAPSED_BNODE,
    BNodeHandling,
    GraphHelper,
    affix_tuples,
    file_uri_to_path,
)


def check(
    file_uri: str,
    expected_windows_path: Optional[str],
    expected_posix_path: Optional[str],
) -> None:
    if expected_windows_path is not None:
        expected_windows_path_object = PureWindowsPath(expected_windows_path)
    if expected_posix_path is not None:
        expected_posix_path_object = PurePosixPath(expected_posix_path)

    if expected_windows_path is not None:
        if os.name == "nt":
            assert file_uri_to_path(file_uri) == expected_windows_path_object
        assert (
            file_uri_to_path(file_uri, PureWindowsPath) == expected_windows_path_object
        )

    if expected_posix_path is not None:
        if os.name != "nt":
            assert file_uri_to_path(file_uri) == expected_posix_path_object
        assert file_uri_to_path(file_uri, PurePosixPath) == expected_posix_path_object


@pytest.mark.parametrize(
    "file_uri,expected_windows_path,expected_posix_path",
    [
        (
            r"file:///C:/Windows/System32/Drivers/etc/hosts",
            r"C:\Windows\System32\Drivers\etc\hosts",
            r"/C:/Windows/System32/Drivers/etc/hosts",
        ),
        (
            r"file:///C%3A/Windows/System32/Drivers/etc/hosts",
            None,
            r"/C:/Windows/System32/Drivers/etc/hosts",
        ),
        (
            r"file:///C:/some%20dir/some%20file",
            r"C:\some dir\some file",
            r"/C:/some dir/some file",
        ),
        (
            r"file:///C%3A/some%20dir/some%20file",
            None,
            r"/C:/some dir/some file",
        ),
        (
            r"file:///C:/Python27/Scripts/pip.exe",
            r"C:\Python27\Scripts\pip.exe",
            r"/C:/Python27/Scripts/pip.exe",
        ),
        (
            r"file:///C:/yikes/paths%20with%20spaces.txt",
            r"C:\yikes\paths with spaces.txt",
            r"/C:/yikes/paths with spaces.txt",
        ),
        (
            r"file://localhost/c:/WINDOWS/clock.avi",
            r"c:\WINDOWS\clock.avi",
            r"/c:/WINDOWS/clock.avi",
        ),
        (r"file:///home/example/.profile", None, r"/home/example/.profile"),
        (r"file:///c|/path/to/file", r"c:\path\to\file", r"/c|/path/to/file"),
        (r"file:/c|/path/to/file", r"c:\path\to\file", r"/c|/path/to/file"),
        (r"file:c|/path/to/file", r"c:\path\to\file", r"c|/path/to/file"),
        (r"file:///c:/path/to/file", r"c:\path\to\file", r"/c:/path/to/file"),
        (r"file:/c:/path/to/file", r"c:\path\to\file", r"/c:/path/to/file"),
        (r"file:c:/path/to/file", r"c:\path\to\file", r"c:/path/to/file"),
        (r"file:/path/to/file", None, r"/path/to/file"),
        (r"file:///home/user/some%20file.txt", None, r"/home/user/some file.txt"),
        (
            r"file:///C:/some%20dir/some%20file.txt",
            r"C:\some dir\some file.txt",
            r"/C:/some dir/some file.txt",
        ),
    ],
)
def test_paths(
    file_uri: str,
    expected_windows_path: Optional[str],
    expected_posix_path: Optional[str],
) -> None:
    check(file_uri, expected_windows_path, expected_posix_path)


@dataclass
class SetsEqualTestCase:
    equal: bool
    format: Union[str, Tuple[str, str]]
    bnode_handling: BNodeHandling
    lhs: str
    rhs: str

    @property
    def lhs_format(self) -> str:
        if isinstance(self.format, tuple):
            return self.format[0]
        return self.format

    @property
    def rhs_format(self) -> str:
        if isinstance(self.format, tuple):
            return self.format[1]
        return self.format


@pytest.mark.parametrize(
    "test_case",
    [
        SetsEqualTestCase(
            equal=False,
            format="turtle",
            bnode_handling=BNodeHandling.COMPARE,
            lhs="""
            @prefix eg: <example:> .
            _:a _:b _:c .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
            rhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
        ),
        SetsEqualTestCase(
            equal=True,
            format="turtle",
            bnode_handling=BNodeHandling.EXCLUDE,
            lhs="""
            @prefix eg: <example:> .
            _:a _:b _:c .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
            rhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
        ),
        SetsEqualTestCase(
            equal=True,
            format="turtle",
            bnode_handling=BNodeHandling.COLLAPSE,
            lhs="""
            @prefix eg: <example:> .
            _:a _:b _:c .
            _:z _:b _:c .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
            rhs=f"""
            @prefix eg: <example:> .
            <{COLLAPSED_BNODE}> <{COLLAPSED_BNODE}> <{COLLAPSED_BNODE}>.
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
        ),
        SetsEqualTestCase(
            equal=True,
            format="turtle",
            bnode_handling=BNodeHandling.COMPARE,
            lhs="""
            <example:o0> <example:p0> <example:s0> .
            <example:o1> <example:p1> <example:s1> .
            """,
            rhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
        ),
        SetsEqualTestCase(
            equal=False,
            format="turtle",
            bnode_handling=BNodeHandling.COMPARE,
            lhs="""
            <example:o0> <example:p0> <example:s0> .
            <example:o1> <example:p1> <example:s1> .
            <example:o2> <example:p2> <example:s2> .
            """,
            rhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            """,
        ),
        SetsEqualTestCase(
            equal=True,
            format=("nquads", "trig"),
            bnode_handling=BNodeHandling.EXCLUDE,
            lhs="""
            <example:o0> <example:p0> <example:s0> .
            <example:o1> <example:p1> <example:s1> .
            <example:o2> <example:p2> <example:s2> .
            """,
            rhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:o1 eg:p1 eg:s1 .
            eg:o2 eg:p2 eg:s2 .
            """,
        ),
        SetsEqualTestCase(
            equal=True,
            format=("nquads", "trig"),
            bnode_handling=BNodeHandling.COMPARE,
            lhs="""
            <example:o0> <example:p0> <example:s0> .
            <example:o1> <example:p1> <example:s1> <example:g1>.
            <example:o2> <example:p2> <example:s2> <example:g2> .
            """,
            rhs="""
            @prefix eg: <example:> .
            eg:o0 eg:p0 eg:s0 .
            eg:g1 {
                eg:o1 eg:p1 eg:s1 .
            }
            eg:g2 {
                eg:o2 eg:p2 eg:s2 .
            }
            """,
        ),
        SetsEqualTestCase(
            equal=True,
            format="n3",
            bnode_handling=BNodeHandling.COMPARE,
            lhs="""
            { <example:ss0> <example:sp0> <example:so0> } <example:p0> {}.
            """,
            rhs="""
            @prefix eg: <example:> .
            { eg:ss0 eg:sp0 eg:so0 } eg:p0 {}.
            """,
        ),
        SetsEqualTestCase(
            equal=True,
            format="n3",
            bnode_handling=BNodeHandling.COMPARE,
            lhs="""
            { <example:ss0> <example:sp0> <example:so0> } <example:p0> {}.
            """,
            rhs="""
            @prefix eg: <example:> .
            { eg:ss0 eg:sp0 eg:so0 } eg:p0 {}.
            """,
        ),
        SetsEqualTestCase(
            equal=True,
            format="n3",
            bnode_handling=BNodeHandling.COMPARE,
            lhs="""
            { { <example:sss0> <example:ssp0> <example:sso0> } <example:sp0> <example:so0> } <example:p0> {}.
            """,
            rhs="""
            @prefix eg: <example:> .
            { { eg:sss0 eg:ssp0 eg:sso0 } eg:sp0 eg:so0 } eg:p0 {}.
            """,
        ),
    ],
)
def test_assert_sets_equal(test_case: SetsEqualTestCase):
    """
    GraphHelper.sets_equals and related functions work correctly in both
    positive and negative cases.
    """
    lhs_graph: Graph = Graph().parse(data=test_case.lhs, format=test_case.lhs_format)
    rhs_graph: Graph = Graph().parse(data=test_case.rhs, format=test_case.rhs_format)

    public_id = URIRef("example:graph")
    lhs_dataset: Dataset = Dataset()
    lhs_dataset.parse(
        data=test_case.lhs, format=test_case.lhs_format, publicID=public_id
    )

    rhs_dataset: Dataset = Dataset()
    rhs_dataset.parse(
        data=test_case.rhs, format=test_case.rhs_format, publicID=public_id
    )

    assert isinstance(lhs_dataset, Dataset)
    assert isinstance(rhs_dataset, Dataset)
    graph: Graph
    cgraph: Dataset
    for graph, cgraph in ((lhs_graph, lhs_dataset), (rhs_graph, rhs_dataset)):
        GraphHelper.assert_sets_equals(graph, graph, BNodeHandling.COLLAPSE)
        GraphHelper.assert_sets_equals(cgraph, cgraph, BNodeHandling.COLLAPSE)
        GraphHelper.assert_triple_sets_equals(graph, graph, BNodeHandling.COLLAPSE)
        GraphHelper.assert_triple_sets_equals(cgraph, cgraph, BNodeHandling.COLLAPSE)
        GraphHelper.assert_quad_sets_equals(cgraph, cgraph, BNodeHandling.COLLAPSE)

    if not test_case.equal:
        with pytest.raises(AssertionError):
            GraphHelper.assert_sets_equals(
                lhs_graph, rhs_graph, test_case.bnode_handling
            )
        with pytest.raises(AssertionError):
            GraphHelper.assert_sets_equals(
                lhs_dataset, rhs_dataset, test_case.bnode_handling
            )
        with pytest.raises(AssertionError):
            GraphHelper.assert_triple_sets_equals(
                lhs_graph, rhs_graph, test_case.bnode_handling
            )
        with pytest.raises(AssertionError):
            GraphHelper.assert_triple_sets_equals(
                lhs_dataset, rhs_dataset, test_case.bnode_handling
            )
        with pytest.raises(AssertionError):
            GraphHelper.assert_quad_sets_equals(
                lhs_dataset, rhs_dataset, test_case.bnode_handling
            )
    else:
        GraphHelper.assert_sets_equals(lhs_graph, rhs_graph, test_case.bnode_handling)
        GraphHelper.assert_sets_equals(
            lhs_dataset, rhs_dataset, test_case.bnode_handling
        )
        GraphHelper.assert_triple_sets_equals(
            lhs_graph, rhs_graph, test_case.bnode_handling
        )
        GraphHelper.assert_triple_sets_equals(
            lhs_dataset, rhs_dataset, test_case.bnode_handling
        )
        GraphHelper.assert_quad_sets_equals(
            lhs_dataset, rhs_dataset, test_case.bnode_handling
        )


@pytest.mark.parametrize(
    ["tuples", "prefix", "suffix", "expected_result"],
    [
        (
            [("a",), ("b",), ("c",)],
            None,
            None,
            [("a",), ("b",), ("c",)],
        ),
        (
            [("a",), ("b",), ("c",)],
            (1, 2),
            None,
            [
                (1, 2, "a"),
                (1, 2, "b"),
                (1, 2, "c"),
            ],
        ),
        (
            [("a",), ("b",), ("c",)],
            None,
            (3, 4),
            [
                ("a", 3, 4),
                ("b", 3, 4),
                ("c", 3, 4),
            ],
        ),
    ],
)
def test_prefix_tuples(
    tuples: List[Tuple[Any, ...]],
    prefix: Tuple[Any, ...],
    suffix: Tuple[Any, ...],
    expected_result: List[Tuple[Any, ...]],
) -> None:
    assert expected_result == list(affix_tuples(prefix, tuples, suffix))


@pytest.mark.parametrize(
    ["graph_type", "format", "lhs", "rhs", "expected_result"],
    [
        (
            Dataset,
            "trig",
            """
            @prefix eg: <example:> .

            _:b0 eg:p0 eg:o0.
            eg:s1 eg:p1 eg:o1.

            eg:g0 {
                _:g0b0 eg:g0p0 eg:g0o0.
                eg:g0s1 eg:g0p1 eg:g0o1.
            }
            """,
            """
            @prefix eg: <example:> .

            _:b1 eg:p0 eg:o0.
            eg:s1 eg:p1 eg:o1.

            eg:g0 {
                _:g0b1 eg:g0p0 eg:g0o0.
                eg:g0s1 eg:g0p1 eg:g0o1.
            }
            """,
            None,
        ),
        (
            Dataset,
            "trig",
            """
            @prefix eg: <example:> .

            eg:g0 {
                _:b0 eg:g0p0 eg:g0o0.
                eg:g0s1 eg:g0p1 eg:g0o1.
            }
            """,
            """
            @prefix eg: <example:> .

            eg:g0 {
                _:b1 eg:g0p0 eg:g0o1.
                eg:g0s1 eg:g0p1 eg:g0o1.
            }
            """,
            AssertionError,
        ),
        (
            Dataset,
            "trig",
            """
            @prefix eg: <example:> .

            eg:g0 {
                _:b0 eg:g0p0 eg:g0o0.
                eg:g0s1 eg:g0p1 eg:g0o1.
            }
            """,
            """
            @prefix eg: <example:> .

            eg:g0 {
                _:b1 eg:g0p0 eg:g0o0.
                eg:g0s1 eg:g0p1 eg:g0o1.
            }

            eg:g1 {
                _:b0 eg:g1p0 eg:g1o0.
                eg:g1s1 eg:g1p1 eg:g1o1.
            }
            """,
            AssertionError,
        ),
    ],
)
def test_assert_cgraph_isomorphic(
    graph_type: Type[ConjunctiveGraph],
    format: str,
    lhs: str,
    rhs: str,
    expected_result: Union[None, Type[Exception]],
) -> None:
    lhs_graph = graph_type()
    lhs_graph.parse(data=lhs, format=format)
    rhs_graph = graph_type()
    rhs_graph.parse(data=rhs, format=format)
    catcher: Optional[pytest.ExceptionInfo[Exception]] = None
    with ExitStack() as xstack:
        if expected_result is not None:
            catcher = xstack.enter_context(pytest.raises(expected_result))
        GraphHelper.assert_cgraph_isomorphic(lhs_graph, rhs_graph, exclude_bnodes=True)
    if expected_result is None:
        assert catcher is None
    else:
        assert catcher is not None



import pytest

from rdflib import Graph
from rdflib.plugins.stores.auditable import AuditableStore
from test.utils.namespace import EGDO


@pytest.fixture
def get_graph():
    g = Graph("Memory")

    g.add((EGDO.s0, EGDO.p0, EGDO.o0))
    g.add((EGDO.s0, EGDO.p0, EGDO.o0bis))

    t = Graph(AuditableStore(g.store), g.identifier)

    yield g, t


def test_add_commit(get_graph):
    g, t = get_graph
    t.add((EGDO.s1, EGDO.p1, EGDO.o1))
    assert set(t) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
            (EGDO.s1, EGDO.p1, EGDO.o1),
        ]
    )

    t.commit()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
            (EGDO.s1, EGDO.p1, EGDO.o1),
        ]
    )


def test_remove_commit(get_graph):
    g, t = get_graph
    t.remove((EGDO.s0, EGDO.p0, EGDO.o0))
    assert set(t) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )
    t.commit()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_multiple_remove_commit(get_graph):
    g, t = get_graph
    t.remove((EGDO.s0, EGDO.p0, None))
    assert set(t) == set([])
    t.commit()
    assert set(g) == set([])


def test_noop_add_commit(get_graph):
    g, t = get_graph
    t.add((EGDO.s0, EGDO.p0, EGDO.o0))
    assert set(t) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )
    t.commit()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_noop_remove_commit(get_graph):
    g, t = get_graph
    t.add((EGDO.s0, EGDO.p0, EGDO.o0))
    assert set(t) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )

    t.commit()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_add_remove_commit(get_graph):
    g, t = get_graph
    t.add((EGDO.s1, EGDO.p1, EGDO.o1))
    t.remove((EGDO.s1, EGDO.p1, EGDO.o1))
    assert set(t) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )
    t.commit()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_remove_add_commit(get_graph):
    g, t = get_graph
    t.remove((EGDO.s1, EGDO.p1, EGDO.o1))
    t.add((EGDO.s1, EGDO.p1, EGDO.o1))
    assert set(t) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
            (EGDO.s1, EGDO.p1, EGDO.o1),
        ]
    )
    t.commit()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
            (EGDO.s1, EGDO.p1, EGDO.o1),
        ]
    )


def test_add_rollback(get_graph):
    g, t = get_graph
    t.add((EGDO.s1, EGDO.p1, EGDO.o1))
    t.rollback()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_remove_rollback(get_graph):
    g, t = get_graph
    t.remove((EGDO.s0, EGDO.p0, EGDO.o0))
    t.rollback()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_multiple_remove_rollback(get_graph):
    g, t = get_graph
    t.remove((EGDO.s0, EGDO.p0, None))
    t.rollback()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_noop_add_rollback(get_graph):
    g, t = get_graph
    t.add((EGDO.s0, EGDO.p0, EGDO.o0))
    t.rollback()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_noop_remove_rollback(get_graph):
    g, t = get_graph
    t.add((EGDO.s0, EGDO.p0, EGDO.o0))
    t.rollback()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_add_remove_rollback(get_graph):
    g, t = get_graph
    t.add((EGDO.s1, EGDO.p1, EGDO.o1))
    t.remove((EGDO.s1, EGDO.p1, EGDO.o1))
    t.rollback()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_remove_add_rollback(get_graph):
    g, t = get_graph
    t.remove((EGDO.s1, EGDO.p1, EGDO.o1))
    t.add((EGDO.s1, EGDO.p1, EGDO.o1))
    t.rollback()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


@pytest.fixture
def get_empty_graph():
    g = Graph("Memory")

    t = Graph(AuditableStore(g.store), g.identifier)

    yield g, t


def test_add_commit_empty(get_empty_graph):
    g, t = get_empty_graph
    t.add((EGDO.s1, EGDO.p1, EGDO.o1))
    assert set(t) == set(
        [
            (EGDO.s1, EGDO.p1, EGDO.o1),
        ]
    )
    t.commit()
    assert set(g) == set(
        [
            (EGDO.s1, EGDO.p1, EGDO.o1),
        ]
    )


def test_add_rollback_empty(get_empty_graph):
    g, t = get_empty_graph
    t.add((EGDO.s1, EGDO.p1, EGDO.o1))
    t.rollback()
    assert set(g) == set([])


@pytest.fixture
def get_concurrent_graph():
    g = Graph("Memory")

    g.add((EGDO.s0, EGDO.p0, EGDO.o0))
    g.add((EGDO.s0, EGDO.p0, EGDO.o0bis))
    t1 = Graph(AuditableStore(g.store), g.identifier)
    t2 = Graph(AuditableStore(g.store), g.identifier)
    t1.add((EGDO.s1, EGDO.p1, EGDO.o1))
    t2.add((EGDO.s2, EGDO.p2, EGDO.o2))
    t1.remove((EGDO.s0, EGDO.p0, EGDO.o0))
    t2.remove((EGDO.s0, EGDO.p0, EGDO.o0bis))

    yield g, t1, t2


def test_commit_commit(get_concurrent_graph):
    g, t1, t2 = get_concurrent_graph
    t1.commit()
    t2.commit()
    assert set(g) == set(
        [
            (EGDO.s1, EGDO.p1, EGDO.o1),
            (EGDO.s2, EGDO.p2, EGDO.o2),
        ]
    )


def test_commit_rollback(get_concurrent_graph):
    g, t1, t2 = get_concurrent_graph
    t1.commit()
    t2.rollback()
    assert set(g) == set(
        [
            (EGDO.s1, EGDO.p1, EGDO.o1),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_rollback_commit(get_concurrent_graph):
    g, t1, t2 = get_concurrent_graph
    t1.rollback()
    t2.commit()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s2, EGDO.p2, EGDO.o2),
        ]
    )


def test_rollback_rollback(get_concurrent_graph):
    g, t1, t2 = get_concurrent_graph
    t1.rollback()
    t2.rollback()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


@pytest.fixture
def get_embedded_graph():
    g = Graph("Memory")

    g.add((EGDO.s0, EGDO.p0, EGDO.o0))
    g.add((EGDO.s0, EGDO.p0, EGDO.o0bis))

    t1 = Graph(AuditableStore(g.store), g.identifier)
    t1.add((EGDO.s1, EGDO.p1, EGDO.o1))
    t1.remove((EGDO.s0, EGDO.p0, EGDO.o0bis))

    t2 = Graph(AuditableStore(t1.store), t1.identifier)
    t2.add((EGDO.s2, EGDO.p2, EGDO.o2))
    t2.remove((EGDO.s1, EGDO.p1, EGDO.o1))

    yield g, t1, t2


def test_commit_commit_embedded(get_embedded_graph):
    g, t1, t2 = get_embedded_graph
    assert set(t2) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s2, EGDO.p2, EGDO.o2),
        ]
    )
    t2.commit()
    assert set(t1) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s2, EGDO.p2, EGDO.o2),
        ]
    )
    t1.commit()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s2, EGDO.p2, EGDO.o2),
        ]
    )


def test_commit_rollback_embedded(get_embedded_graph):
    g, t1, t2 = get_embedded_graph
    t2.commit()
    t1.rollback()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )


def test_rollback_commit_embedded(get_embedded_graph):
    g, t1, t2 = get_embedded_graph
    t2.rollback()
    assert set(t1) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s1, EGDO.p1, EGDO.o1),
        ]
    )
    t1.commit()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s1, EGDO.p1, EGDO.o1),
        ]
    )


def test_rollback_rollback_embedded(get_embedded_graph):
    g, t1, t2 = get_embedded_graph
    t2.rollback()
    t1.rollback()
    assert set(g) == set(
        [
            (EGDO.s0, EGDO.p0, EGDO.o0),
            (EGDO.s0, EGDO.p0, EGDO.o0bis),
        ]
    )



from __future__ import annotations

import enum
import itertools
import logging
import re
import socket
from contextlib import ExitStack
from dataclasses import dataclass, field
from functools import lru_cache
from pathlib import Path, PosixPath, PurePath
from typing import (
    IO,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    TextIO,
    Tuple,
    Union,
    cast,
)
from urllib.parse import urlsplit, urlunsplit

import pytest
from _pytest.mark.structures import Mark, MarkDecorator, ParameterSet

import rdflib
import rdflib.plugin
from rdflib import RDF, XSD, Graph, Literal, Namespace, URIRef
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID, ConjunctiveGraph, Dataset
from rdflib.serializer import Serializer
from test.utils import GraphHelper, get_unique_plugins
from test.utils.destination import DestinationType, DestParmType, DestRef
from test.utils.namespace import EGDC, EGSCHEME, EGURN


@pytest.mark.parametrize(
    "format, tuple_index, is_keyword",
    [
        (format, tuple_index, keyword)
        for format, (tuple_index, keyword) in itertools.product(
            ["turtle", "n3", "trig"],
            [
                (0, False),
                (1, True),
                (2, False),
            ],
        )
    ]
    + [("trig", 3, False)],
)
def test_rdf_type(format: str, tuple_index: int, is_keyword: bool) -> None:
    NS = Namespace("example:")  # noqa: N806
    graph = ConjunctiveGraph()
    graph.bind("eg", NS)
    nodes = [NS.subj, NS.pred, NS.obj, NS.graph]
    nodes[tuple_index] = RDF.type
    quad = cast(Tuple[URIRef, URIRef, URIRef, URIRef], tuple(nodes))
    # type error: Argument 1 to "add" of "ConjunctiveGraph" has incompatible type "Tuple[URIRef, URIRef, URIRef, URIRef]"; expected "Union[Tuple[Node, Node, Node], Tuple[Node, Node, Node, Optional[Graph]]]"
    graph.add(quad)  # type: ignore[arg-type]
    data = graph.serialize(format=format)
    logging.info("data = %s", data)
    assert NS in data
    if is_keyword:
        assert str(RDF) not in data
    else:
        assert str(RDF) in data
    parsed_graph = ConjunctiveGraph()
    parsed_graph.parse(data=data, format=format)
    GraphHelper.assert_triple_sets_equals(graph, parsed_graph)


EG = Namespace("example:")


@pytest.fixture(scope="session")
def simple_graph() -> Graph:
    """
    This is a simple graph with no BNodes that can be used in tests.
    Assumptions/assertions should not be made about the triples in it, other
    than that it contains no blank nodes.
    """
    graph = Graph()
    graph.add((EGSCHEME.subject, EGSCHEME.predicate, EGSCHEME.object))
    graph.add((EGSCHEME.subject, EGSCHEME.predicate, Literal(12)))
    graph.add(
        (
            EGDC.subject,
            EGDC.predicate,
            Literal("日本語の表記体系", lang="jpx"),
        )
    )
    graph.add((EGURN.subject, EGSCHEME.predicate, EGSCHEME.subject))
    graph.add(
        (EGSCHEME.object, EGDC.predicate, Literal("XSD string", datatype=XSD.string))
    )
    return graph


@pytest.fixture(scope="session")
def simple_dataset() -> Dataset:
    """
    This is a simple dataset with no BNodes that can be used in tests.
    Assumptions/assertions should not be made about the quads in it, other
    than that it contains no blank nodes.
    """
    graph = Dataset()
    graph.default_context.add((EGSCHEME.subject, EGSCHEME.predicate, EGSCHEME.object))
    graph.default_context.add((EGURN.subject, EGURN.predicate, EGURN.object))
    graph.default_context.add((EGDC.subject, EGDC.predicate, Literal("typeless")))
    graph.get_context(EGSCHEME.graph).add(
        (EGSCHEME.subject, EGSCHEME.predicate, EGSCHEME.object)
    )
    graph.get_context(EGSCHEME.graph).add(
        (EGSCHEME.subject, EGSCHEME.predicate, Literal(12))
    )
    graph.get_context(EGSCHEME.graph).add(
        (
            EGDC.subject,
            EGDC.predicate,
            Literal("日本語の表記体系", lang="jpx"),
        )
    )
    graph.get_context(EGSCHEME.graph).add(
        (EGURN.subject, EGSCHEME.predicate, EGSCHEME.subject)
    )
    graph.get_context(EGURN.graph).add(
        (EGSCHEME.subject, EGSCHEME.predicate, EGSCHEME.object)
    )
    graph.get_context(EGURN.graph).add((EGSCHEME.subject, EGDC.predicate, EGDC.object))
    graph.get_context(EGURN.graph).add(
        (EGSCHEME.subject, EGDC.predicate, Literal("XSD string", datatype=XSD.string))
    )
    return graph


def test_serialize_to_purepath(tmp_path: Path, simple_graph: Graph):
    tfpath = tmp_path / "out.nt"
    simple_graph.serialize(destination=tfpath, format="nt", encoding="utf-8")
    graph_check = Graph()
    graph_check.parse(source=tfpath, format="nt")

    GraphHelper.assert_triple_sets_equals(simple_graph, graph_check)


def test_serialize_to_path(tmp_path: Path, simple_graph: Graph):
    tfpath = tmp_path / "out.nt"
    simple_graph.serialize(destination=tfpath, format="nt", encoding="utf-8")
    graph_check = Graph()
    graph_check.parse(source=tfpath, format="nt")

    GraphHelper.assert_triple_sets_equals(simple_graph, graph_check)


def test_serialize_to_neturl(simple_graph: Graph):
    with pytest.raises(Exception) as raised:
        simple_graph.serialize(
            destination="http://example.com/", format="nt", encoding="utf-8"
        )
    assert isinstance(
        raised.value,
        (
            FileNotFoundError,  # This is for Posix
            OSError,  # This is for Windows
        ),
    )


def test_serialize_to_fileurl(tmp_path: Path, simple_graph: Graph):
    tfpath = tmp_path / "out.nt"
    tfurl = tfpath.as_uri()
    assert re.match(r"^file:", tfurl)
    assert not tfpath.exists()
    simple_graph.serialize(destination=tfurl, format="nt", encoding="utf-8")
    assert tfpath.exists()
    graph_check = Graph()
    graph_check.parse(source=tfpath, format="nt")
    GraphHelper.assert_triple_sets_equals(simple_graph, graph_check)


def test_serialize_badformat(simple_graph: Graph) -> None:
    # NOTE: This is using an fully qualified name because of reloading that
    # occurs in `test_plugins.py` which results in the the type identity being
    # different.
    with pytest.raises(rdflib.plugin.PluginException) as ctx:
        simple_graph.serialize(destination="http://example.com/", format="badformat")
    assert "badformat" in f"{ctx.value}"


DESTINATION_TYPES = {
    DestinationType.RETURN,
    DestinationType.PATH,
    DestinationType.PURE_PATH,
    DestinationType.STR_PATH,
    DestinationType.FILE_URI,
    DestinationType.BINARY_IO,
}


GraphDestParamType = Union[Path, PurePath, str, IO[bytes]]


def narrow_dest_param(param: DestParmType) -> GraphDestParamType:
    assert not (hasattr(param, "write") and hasattr(param, "encoding"))
    assert not isinstance(param, TextIO)
    return param


class GraphType(str, enum.Enum):
    QUAD = enum.auto()
    TRIPLE = enum.auto()


class GraphFormat(str, enum.Enum):
    TRIG = "trig"
    NQUADS = "nquads"
    TRIX = "trix"
    JSON_LD = "json-ld"
    TURTLE = "turtle"
    NT11 = "nt11"
    XML = "xml"
    N3 = "n3"
    NTRIPLES = "ntriples"
    HEXT = "hext"

    @classmethod
    @lru_cache(maxsize=None)
    def info_dict(cls) -> GraphFormatInfoDict:
        return GraphFormatInfoDict.make(
            GraphFormatInfo(
                GraphFormat.TRIG,
                graph_types={GraphType.TRIPLE, GraphType.QUAD},
                encodings={"utf-8"},
            ),
            GraphFormatInfo(
                GraphFormat.NQUADS,
                # TODO FIXME: Currently nquads rejects requests to seralize
                # non-context-aware stores, this does not make a lot of sense and
                # should be fixed.
                #
                graph_types={GraphType.QUAD},
                encodings={"utf-8"},
            ),
            GraphFormatInfo(
                GraphFormat.TRIX,
                # TODO FIXME: Currently trix rejects requests to seralize
                # non-context-aware stores, this does not make a lot of sense and
                # should be fixed.
                #
                graph_types={GraphType.QUAD},
                encodings={"utf-8"},
            ),
            GraphFormatInfo(
                GraphFormat.JSON_LD,
                graph_types={GraphType.TRIPLE, GraphType.QUAD},
                encodings={"utf-8"},
            ),
            GraphFormatInfo(
                GraphFormat.TURTLE,
                graph_types={GraphType.TRIPLE},
                encodings={"utf-8"},
                serializer_list=["longturtle", "turtle"],
            ),
            GraphFormatInfo(
                GraphFormat.NT11,
                graph_types={GraphType.TRIPLE},
                encodings={"utf-8"},
            ),
            GraphFormatInfo(
                GraphFormat.XML,
                graph_types={GraphType.TRIPLE},
                encodings={"utf-8"},
                serializer_list=["xml", "pretty-xml"],
            ),
            GraphFormatInfo(
                GraphFormat.N3,
                graph_types={GraphType.TRIPLE},
                encodings={"utf-8"},
            ),
            GraphFormatInfo(
                GraphFormat.NTRIPLES,
                graph_types={GraphType.TRIPLE},
                encodings={"utf-8"},
            ),
            GraphFormatInfo(
                GraphFormat.HEXT,
                graph_types={GraphType.TRIPLE, GraphType.QUAD},
                encodings={"utf-8"},
            ),
        )

    @property
    def info(self) -> GraphFormatInfo:
        return self.info_dict()[self]

    @classmethod
    @lru_cache(maxsize=None)
    def set(cls) -> Set[GraphFormat]:
        return set(*cls)


@dataclass
class GraphFormatInfo:
    name: GraphFormat
    graph_types: Set[GraphType]
    encodings: Set[str]
    serializer_list: Optional[List[str]] = field(
        default=None, repr=False, hash=False, compare=False
    )
    deserializer_list: Optional[List[str]] = field(
        default=None, repr=False, hash=False, compare=False
    )
    serializers: List[str] = field(default_factory=list, init=False)
    deserializers: List[str] = field(default_factory=list, init=False)

    def __post_init__(self) -> None:
        self.serializers = (
            [self.name.value] if self.serializer_list is None else self.serializer_list
        )
        self.deserializers = (
            [self.name.value]
            if self.deserializer_list is None
            else self.deserializer_list
        )

    @property
    def serializer(self) -> str:
        if not self.serializers:
            raise RuntimeError("no serializers for {self.name}")
        return self.serializers[0]

    @property
    def deserializer(self) -> str:
        if not self.deserializers:
            raise RuntimeError("no deserializer for {self.name}")
        return self.deserializer[0]


class GraphFormatInfoDict(Dict[str, GraphFormatInfo]):
    @classmethod
    def make(cls, *graph_format: GraphFormatInfo) -> GraphFormatInfoDict:
        result = cls()
        for item in graph_format:
            result[item.name] = item
        return result

    def serdes_dict(self) -> Tuple[Dict[str, GraphFormat], Dict[str, GraphFormat]]:
        serializer_dict: Dict[str, GraphFormat] = {}
        deserializer_dict: Dict[str, GraphFormat] = {}
        for format in self.values():
            for serializer in format.serializers:
                serializer_dict[serializer] = format.name
            for deserializer in format.deserializers:
                deserializer_dict[deserializer] = format.name
        return (serializer_dict, deserializer_dict)


serializer_dict, deserializer_dict = GraphFormat.info_dict().serdes_dict()


def test_all_serializers_specified() -> None:
    plugins = get_unique_plugins(Serializer)
    for plugin_refs in plugins.values():
        names = {plugin_ref.name for plugin_ref in plugin_refs}
        assert (
            set(serializer_dict.keys()) != set()
        ), f"serializers does not include any of {names}"


def make_serialize_parse_tests() -> Generator[ParameterSet, None, None]:
    """
    This function generates test parameters for test_serialize_parse.
    """
    xfails: Dict[
        Tuple[str, GraphType, DestinationType, Optional[str]],
        Union[MarkDecorator, Mark],
    ] = {}
    for serializer_name, destination_type in itertools.product(
        serializer_dict.keys(), DESTINATION_TYPES
    ):
        format = serializer_dict[serializer_name]
        encodings: Set[Optional[str]] = {*format.info.encodings, None}
        for encoding, graph_type in itertools.product(
            encodings, format.info.graph_types
        ):
            xfail = xfails.get(
                (serializer_name, graph_type, destination_type, encoding)
            )
            if not xfail:
                if serializer_name in ("trig") and graph_type is GraphType.TRIPLE:
                    xfail = pytest.mark.xfail(
                        raises=AssertionError,
                        reason="""
        TriG serializes non-context aware stores incorrectly, adding a blank
        node graph name which breaks round tripping.
                        """,
                    )
                if serializer_name in ("json-ld"):
                    xfail = pytest.mark.xfail(
                        raises=AssertionError,
                        reason="""
    JSON-LD is dropping datatype:
    -   rdflib.term.Literal('XSD string'),
    +   rdflib.term.Literal('XSD string', datatype=rdflib.term.URIRef('http://www.w3.org/2001/XMLSchema#string')),
                        """,
                    )
                elif serializer_name in ("hext") and graph_type is GraphType.QUAD:
                    xfail = pytest.mark.xfail(
                        raises=AssertionError,
                        reason="""
    hext is injecting datatype:
    -   rdflib.term.Literal('typeless', datatype=rdflib.term.URIRef('http://www.w3.org/2001/XMLSchema#string')),
    +   rdflib.term.Literal('typeless'),
    """,
                    )
            marks = (xfail,) if xfail is not None else ()
            yield pytest.param(
                (serializer_name, graph_type, destination_type, encoding),
                id=f"{serializer_name}-{graph_type.name}-{destination_type.name}-{encoding}",
                marks=marks,
            )


SerializeResultType = Union[bytes, str, Graph]


@pytest.mark.parametrize(
    ["args"],
    make_serialize_parse_tests(),
)
def test_serialize_parse(
    tmp_path: Path,
    simple_graph: Graph,
    simple_dataset: Dataset,
    args: Tuple[str, GraphType, DestinationType, Optional[str]],
) -> None:
    """
    Serialization works correctly with the given arguments and generates output
    that can be parsed to a graph that is identical to the original graph.
    """
    serializer_name, graph_type, destination_type, encoding = args
    format = serializer_dict[serializer_name]
    graph: Union[Graph, Dataset]
    if graph_type == GraphType.QUAD:
        graph = simple_dataset
    elif graph_type == GraphType.TRIPLE:
        graph = simple_graph
    else:
        raise ValueError(f"graph_type {graph_type!r} is not supported")
    with destination_type.make_ref(tmp_path) as dest_ref:
        destination = None if dest_ref is None else narrow_dest_param(dest_ref.param)
        serialize_result = graph.serialize(
            destination=destination,
            format=serializer_name,
            encoding=encoding,
        )

    logging.debug("serialize_result = %r, dest_ref = %s", serialize_result, dest_ref)

    if dest_ref is None:
        if encoding is None:
            assert isinstance(serialize_result, str)
            serialized_data = serialize_result
        else:
            assert isinstance(serialize_result, bytes)
            serialized_data = serialize_result.decode(encoding)
    else:
        assert isinstance(serialize_result, Graph)
        assert dest_ref.path.exists()
        serialized_data = dest_ref.path.read_bytes().decode(
            "utf-8" if encoding is None else encoding
        )

    logging.debug("serialized_data = %s", serialized_data)
    check_serialized(format, graph, serialized_data)


def check_serialized(format: GraphFormat, graph: Graph, data: str) -> None:
    graph_check = Dataset() if isinstance(graph, Dataset) else Graph()
    if not format.info.deserializers:
        return
    graph_check.parse(
        data=data,
        format=format.info.deserializers[0],
        publicID=DATASET_DEFAULT_GRAPH_ID,
    )
    logging.debug("graph = %r, graph_check = %r", graph, graph_check)
    GraphHelper.assert_sets_equals(graph, graph_check)


@dataclass
class SerializeArgs:
    format: str
    opt_dest_ref: Optional[DestRef]

    @property
    def dest_ref(self) -> DestRef:
        if self.opt_dest_ref is None:
            raise RuntimeError("dest_ref is None")
        return self.opt_dest_ref

    @property
    def dest_param(self) -> GraphDestParamType:
        if self.opt_dest_ref is None:
            raise RuntimeError("dest_ref is None")
        return narrow_dest_param(self.opt_dest_ref.param)


SerializeFunctionType = Callable[[Graph, SerializeArgs], SerializeResultType]
StrSerializeFunctionType = Callable[[Graph, SerializeArgs], str]
BytesSerializeFunctionType = Callable[[Graph, SerializeArgs], bytes]
FileSerializeFunctionType = Callable[[Graph, SerializeArgs], Graph]


str_serialize_functions: List[StrSerializeFunctionType] = [
    lambda graph, args: graph.serialize(),
    lambda graph, args: graph.serialize(None),
    lambda graph, args: graph.serialize(None, args.format),
    lambda graph, args: graph.serialize(None, args.format, encoding=None),
    lambda graph, args: graph.serialize(None, args.format, None, None),
    lambda graph, args: graph.serialize(None, format=args.format),
    lambda graph, args: graph.serialize(None, format=args.format, encoding=None),
]


bytes_serialize_functions: List[BytesSerializeFunctionType] = [
    lambda graph, args: graph.serialize(encoding="utf-8"),
    lambda graph, args: graph.serialize(None, args.format, encoding="utf-8"),
    lambda graph, args: graph.serialize(None, args.format, None, "utf-8"),
    lambda graph, args: graph.serialize(None, args.format, None, encoding="utf-8"),
    lambda graph, args: graph.serialize(None, args.format, encoding="utf-8"),
    lambda graph, args: graph.serialize(None, format=args.format, encoding="utf-8"),
]


file_serialize_functions: List[FileSerializeFunctionType] = [
    lambda graph, args: graph.serialize(args.dest_param),
    lambda graph, args: graph.serialize(args.dest_param, encoding=None),
    lambda graph, args: graph.serialize(args.dest_param, encoding="utf-8"),
    lambda graph, args: graph.serialize(args.dest_param, args.format),
    lambda graph, args: graph.serialize(args.dest_param, args.format, encoding=None),
    lambda graph, args: graph.serialize(args.dest_param, args.format, None, None),
    lambda graph, args: graph.serialize(args.dest_param, args.format, encoding="utf-8"),
    lambda graph, args: graph.serialize(
        args.dest_param, args.format, None, encoding="utf-8"
    ),
    lambda graph, args: graph.serialize(args.dest_param, args.format, None, "utf-8"),
]


@pytest.mark.parametrize(
    ["destination_type", "serialize_function"],
    [
        *[
            (DestinationType.RETURN, str_serialize_function)
            for str_serialize_function in str_serialize_functions
        ],
        *[
            (DestinationType.RETURN, bytes_serialize_function)
            for bytes_serialize_function in bytes_serialize_functions
        ],
        *itertools.product(
            {
                DestinationType.BINARY_IO,
                DestinationType.PATH,
                DestinationType.PURE_PATH,
                DestinationType.STR_PATH,
            },
            file_serialize_functions,
        ),
    ],
)
def test_serialize_overloads(
    tmp_path: Path,
    simple_graph: Graph,
    destination_type: DestinationType,
    serialize_function: SerializeFunctionType,
) -> None:
    format = GraphFormat.TURTLE

    with destination_type.make_ref(tmp_path) as dest_ref:
        serialize_result = serialize_function(
            simple_graph, SerializeArgs(format, dest_ref)
        )

    if isinstance(serialize_result, str):
        serialized_data = serialize_result
    elif isinstance(serialize_result, bytes):
        serialized_data = serialize_result.decode("utf-8")
    else:
        assert isinstance(serialize_result, Graph)
        assert dest_ref is not None
        assert dest_ref.path.exists()
        serialized_data = dest_ref.path.read_text(encoding="utf-8")

    check_serialized(format, simple_graph, serialized_data)


def make_test_serialize_to_strdest_tests() -> Generator[ParameterSet, None, None]:
    destination_types: Set[DestinationType] = {
        DestinationType.FILE_URI,
        DestinationType.STR_PATH,
    }
    name_prefixes = [
        r"a_b-",
        r"a%b-",
        r"a%20b-",
        r"a b-",
        r"a b-",
        r"a@b",
        r"a$b",
        r"a!b",
    ]
    if isinstance(Path.cwd(), PosixPath):
        # not valid on windows https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file#naming-conventions
        name_prefixes.extend(
            [
                r"a:b-",
                r"a|b",
            ]
        )
    for destination_type, name_prefix in itertools.product(
        destination_types, name_prefixes
    ):
        yield pytest.param(
            destination_type,
            name_prefix,
            id=f"{destination_type.name}-{name_prefix}",
        )


@pytest.mark.parametrize(
    ["destination_type", "name_prefix"],
    make_test_serialize_to_strdest_tests(),
)
def test_serialize_to_strdest(
    tmp_path: Path,
    simple_graph: Graph,
    destination_type: DestinationType,
    name_prefix: str,
) -> None:
    """
    Serialization works correctly with the given arguments and generates output
    that can be parsed to a graph that is identical to the original graph.
    """
    format = GraphFormat.TURTLE
    encoding = "utf-8"

    def path_factory(
        tmp_path: Path, type: DestinationType, encoding: Optional[str]
    ) -> Path:
        return tmp_path / f"{name_prefix}file-{type.name}-{encoding}"

    with destination_type.make_ref(
        tmp_path,
        encoding=encoding,
        path_factory=path_factory,
    ) as dest_ref:
        assert dest_ref is not None
        destination = narrow_dest_param(dest_ref.param)
        serialize_result = simple_graph.serialize(
            destination=destination,
            format=format.info.serializer,
            encoding=encoding,
        )

    logging.debug("serialize_result = %r, dest_ref = %s", serialize_result, dest_ref)

    assert isinstance(serialize_result, Graph)
    assert dest_ref.path.exists()
    serialized_data = dest_ref.path.read_bytes().decode(
        "utf-8" if encoding is None else encoding
    )

    logging.debug("serialized_data = %s", serialized_data)
    check_serialized(format, simple_graph, serialized_data)


@pytest.mark.parametrize(
    ["authority"],
    [
        ("localhost",),
        ("127.0.0.1",),
        ("example.com",),
        (socket.gethostname(),),
        (socket.getfqdn(),),
    ],
)
def test_serialize_to_fileuri_with_authortiy(
    tmp_path: Path,
    simple_graph: Graph,
    authority: str,
) -> None:
    """
    Serializing to a file URI with authority raises an error.
    """
    destination_type = DestinationType.FILE_URI
    format = GraphFormat.TURTLE

    with ExitStack() as exit_stack:
        dest_ref = exit_stack.enter_context(
            destination_type.make_ref(
                tmp_path,
            )
        )
        assert dest_ref is not None
        assert isinstance(dest_ref.param, str)
        urlparts = urlsplit(dest_ref.param)._replace(netloc=authority)
        use_url = urlunsplit(urlparts)
        logging.debug("use_url = %s", use_url)
        catcher = exit_stack.enter_context(pytest.raises(ValueError))
        simple_graph.serialize(
            destination=use_url,
            format=format.info.serializer,
        )
        assert False  # this should never happen as serialize should always fail
    # type error, mypy thinks this line is unreachable, but it works fine
    assert catcher.value is not None  # type: ignore[unreachable, unused-ignore]



#!/usr/bin/env python
"""
This is a tool that can be used with git difftool to generate a diff that
ignores type hints and comments.

The name of this script, ``diffrtpy`` is short for "diff runtime python", as
this will only compare the parts of the python code that has a runtime impact.

This is to make it easier to review PRs that contain type hints.

To use this script

.. code-block:: bash
    task run -- python -m pip install --upgrade strip-hints black python-minifier
    PYLOGGING_LEVEL=INFO task run -- git difftool -y -x $(readlink -f devtools/diffrtpy.py) upstream/main | tee /var/tmp/compact.diff

Then attach ``/var/tmp/compact.diff`` to the PR.
"""

from __future__ import annotations

import argparse
import logging
import os
import sys
from dataclasses import dataclass, field
from difflib import unified_diff
from pathlib import Path
from typing import List

import black
import python_minifier
from strip_hints import strip_string_to_string


def clean_python(input: Path) -> str:
    code = input.read_text()
    try:
        code = strip_string_to_string(code, to_empty=True, strip_nl=True)
    except Exception:
        logging.warning(
            "failed to strip type hints from %s, falling back to using with type hints",
            input,
        )
        code = code
    code = python_minifier.minify(
        code,
        remove_annotations=True,
        remove_pass=False,
        remove_literal_statements=True,
        combine_imports=False,
        hoist_literals=False,
        rename_locals=False,
        rename_globals=False,
        remove_object_base=False,
        convert_posargs_to_args=False,
        preserve_shebang=True,
    )
    code = black.format_str(code, mode=black.FileMode())
    return code


@dataclass
class Application:
    parser: argparse.ArgumentParser = field(
        default_factory=lambda: argparse.ArgumentParser(add_help=True)
    )

    def __post_init__(self) -> None:
        parser = self.parser
        parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            dest="verbosity",
            help="increase verbosity level",
        )
        parser.add_argument("lhs_file", nargs=1, type=str)
        parser.add_argument("rhs_file", nargs=1, type=str)
        parser.set_defaults(handler=self.handle)

    def run(self, args: List[str]) -> None:
        parse_result = self.parser.parse_args(args)

        verbosity = parse_result.verbosity
        if verbosity is not None:
            root_logger = logging.getLogger("")
            root_logger.propagate = True
            new_level = (
                root_logger.getEffectiveLevel()
                - (min(1, verbosity)) * 10
                - min(max(0, verbosity - 1), 9) * 1
            )
            root_logger.setLevel(new_level)

        logging.debug(
            "sys.executable = %s, args = %s, parse_result = %s, logging.level = %s",
            sys.executable,
            args,
            parse_result,
            logging.getLogger("").getEffectiveLevel(),
        )

        parse_result.handler(parse_result)

    def handle(self, parse_result: argparse.Namespace) -> None:
        logging.debug("entry ...")

        base = os.environ["BASE"]

        lhs_file: Path = Path(parse_result.lhs_file[0])
        rhs_file: Path = Path(parse_result.rhs_file[0])

        logging.debug(
            "base = %s, lhs_file = %s, rhs_file = %s", base, lhs_file, rhs_file
        )

        if lhs_file.name.endswith(".py") and rhs_file.name.endswith(".py"):
            lhs_file_content = clean_python(lhs_file)
            rhs_file_content = clean_python(rhs_file)
        else:
            lhs_file_content = lhs_file.read_text()
            rhs_file_content = rhs_file.read_text()

        lhs_file_lines = lhs_file_content.splitlines(keepends=True)
        rhs_file_lines = rhs_file_content.splitlines(keepends=True)

        sys.stdout.writelines(
            unified_diff(lhs_file_lines, rhs_file_lines, f"a/{base}", f"b/{base}", n=5)
        )


def main() -> None:
    logging.basicConfig(
        level=os.environ.get("PYLOGGING_LEVEL", logging.INFO),
        stream=sys.stderr,
        datefmt="%Y-%m-%dT%H:%M:%S",
        format=(
            "%(asctime)s.%(msecs)03d %(process)d %(thread)d %(levelno)03d:%(levelname)-8s "
            "%(name)-12s %(module)s:%(lineno)s:%(funcName)s %(message)s"
        ),
    )

    Application().run(sys.argv[1:])


if __name__ == "__main__":
    main()



from rdflib import Dataset


def main():
    # RDF patch data
    add_patch = """
    TX .
    A _:bn1 <http://example.org/predicate1> "object1" .
    A _:bn1 <http://example.org/predicate2> "object2" .
    TC .
    """

    delete_patch = """
    TX .
    D _:bn1 <http://example.org/predicate1> "object1" .
    TC .
    """

    ds = Dataset()

    # Apply add patch
    ds.parse(data=add_patch, format="patch")
    print("After add patch:")
    for triple in ds:
        print(triple)

    # Apply delete patch
    ds.parse(data=delete_patch, format="patch")
    print("After delete patch:")
    for triple in ds:
        print(triple)


if __name__ == "__main__":
    main()



from rdflib import Graph
from test.data import BOB, CHEESE, HATES, LIKES, MICHEL, PIZZA, TAREK


class TestGraphSlice:
    def test_slice(self):
        """
        We pervert the slice object,
        and use start, stop, step as subject, predicate, object

        all operations return generators over full triples
        """

        def sl(x, y):
            return len(list(x)) == y

        def soe(x, y):
            return set([a[2] for a in x]) == set(y)  # equals objects

        g = Graph()
        g.add((TAREK, LIKES, PIZZA))
        g.add((TAREK, LIKES, CHEESE))
        g.add((MICHEL, LIKES, PIZZA))
        g.add((MICHEL, LIKES, CHEESE))
        g.add((BOB, LIKES, CHEESE))
        g.add((BOB, HATES, PIZZA))
        g.add((BOB, HATES, MICHEL))  # gasp!

        # Single terms are all trivial:

        # single index slices by subject, i.e. return triples((x,None,None))
        # tell me everything about "TAREK"
        sl(g[TAREK], 2)

        # single slice slices by s,p,o, with : used to split
        # tell me everything about "TAREK" (same as above)
        sl(g[TAREK::], 2)

        # give me every "LIKES" relationship
        sl(g[:LIKES:], 5)

        # give me every relationship to PIZZA
        sl(g[::PIZZA], 3)

        # give me everyone who LIKES PIZZA
        sl(g[:LIKES:PIZZA], 2)

        # does TAREK like PIZZA?
        assert g[TAREK:LIKES:PIZZA] is True

        # More intesting is using paths

        # everything hated or liked
        sl(g[: HATES | LIKES], 7)
